import re
from functools import cached_property
from typing import Optional, Literal, Any, Union

from pydantic import BaseModel, Field, model_validator

RE_FLAGS = list[
    Literal[
        "IGNORECASE", "MULTILINE", "DOTALL", "VERBOSE", "UNICODE", "ASCII", "LOCALE", "I", "M", "S", "X", "A", "U", "L"
    ]
]


class RegexValue(BaseModel):
    pattern: str = Field(
        description="The regex pattern to use for the Attribute Value.", title="Regex Pattern", examples=["^(\\w\\w)-"]
    )
    group: int = Field(0, description="The group number to extract from the regex match.", title="Regex Group", ge=0)
    flags: Optional[RE_FLAGS] = Field(
        default_factory=list,
        description="Regex flags to use for the regex pattern.",
        title="Regex Flags",
        examples=["IGNORECASE", "MULTILINE", "DOTALL", "VERBOSE", "UNICODE", "ASCII", "LOCALE"],
    )

    @cached_property
    def compiled(self):
        """Compile the regex pattern for use in matching."""
        flags = 0
        for _ in self.flags:
            flags |= getattr(re, _)
        return re.compile(self.pattern, flags=flags)


class BaseValue(BaseModel):
    static: Optional[Any] = Field(
        None,
        description="A static value to assign to the Attribute.",
        title="Static Value",
        examples=["PROD", "TEST", "LAB"],
    )
    regex: Optional[RegexValue] = Field(
        None, description="A regex pattern to extract a value from the returned data.", title="Regex Value"
    )
    transform: Optional[Literal["upper", "lower"]] = Field(
        None,
        description="Transform the value to upper or lower case (not used for static values).",
        title="Transformation",
        examples=["upper", "lower"],
    )
    mapping: dict[str, Any] = Field(
        default_factory=dict,
        description="A dictionary mapping the returned data or regex group to a static value to assign to the "
        "Attribute. Any transformations is applied prior to lookup.",
        title="Attribute Value Mapping",
        examples=[{"US": "United States", "usnyc": "New York City"}],
    )
    null_value: Optional[Any] = Field(
        None,
        description="Value to assign if the column is null or no regex match.",
        title="Null Value",
        examples=["unknown", "N/A"],
    )
    default_mapping_value: Optional[Any] = Field(
        None,
        description="Default Value to assign in conjunction with mapping if no key/value pair is found.",
        title="Default Mapping Value",
        examples=["unknown", "N/A"],
    )

    def calc_mapped_value(self, value: str) -> tuple[Union[str, None], str, Union[dict[str, str], None]]:
        regex_match = value if self.regex else None
        if not self.mapping:
            return value, regex_match, None
        if value in self.mapping:
            # If a mapping is set, we want to look up the value in the mapping.
            mapped = str(self.mapping.get(value))
            return mapped, regex_match, {value: mapped}
        if self.default_mapping_value:
            # If the value is not in the mapping, we want to return the default mapping value if set.
            mapped = str(self.default_mapping_value)
            return mapped, regex_match, {value: mapped}
        # If the value is not in the mapping and no default mapping value is set, we return None.
        return None, regex_match, None


class ConfigValue(BaseValue, BaseModel):
    config: Literal["current", "startup", "log"] = Field(
        "current",
        description="The configuration to use for the Attribute Value. Not all vendors and families support snapshot "
        "configuration tasks and some do not startup configs.",
        title="Configuration Type",
        examples=["current", "startup"],
    )
    static: Optional[Any] = Field(
        None,
        description="A static value to assign to the Attribute if a regex match is found in the configuration.",
        title="Static Value",
        examples=["PROD", "TEST", "LAB"],
    )
    regex: RegexValue = Field(
        description="A regex pattern to search the config; if a static value is used then the group is ignored.",
        title="Regex Value",
    )
    no_config_value: Optional[str] = Field(
        None,
        description="Value to assign if the config is not found or supported.",
        title="No Config Value",
        examples=["unknown", "N/A"],
    )
    no_log_value: Optional[str] = Field(
        None,
        description="Value to assign if the log file is not found or supported.",
        title="No Log Value",
        examples=["unknown", "N/A"],
    )

    @model_validator(mode="after")
    def _validate(self):
        if self.default_mapping_value and not self.mapping:
            raise ValueError("'default_mapping_value' must be used with 'mapping'.")
        if self.mapping and self.static:
            raise ValueError("Mapping cannot be used with a Static Value.")
        return self

    def calc_value(self, data: str) -> tuple[Union[str, None], Union[str, None], Union[dict[str, str], None]]:
        if self.transform:
            # If a transform is set, we want to apply it to the value.
            data = data.upper() if self.transform == "upper" else data.lower()
        match = self.regex.compiled.search(data)
        if match and self.static:
            # If the static value is set, we want to return it directly on the matched config.
            return str(self.static), match.group(0), None
        if not match or self.regex.group > len(match.groups()):
            return self.null_value, None, None
        value = match.group(self.regex.group)
        return self.calc_mapped_value(value)


class Sort(BaseModel):
    column: str = Field(
        description="The name of the column to sort by.",
        title="API Column Name",
        examples=["primaryIp", "hostname"],
    )
    order: Literal["asc", "desc"] = Field(
        "asc",
        description="The order to sort the results in.",
        title="Table Sort",
        examples=["asc", "desc"],
    )


class TableValue(BaseValue, BaseModel):
    api_endpoint: str = Field(
        description="The IP Fabric API table endpoint to fetch data from.",
        title="API Table Endpoint",
        examples=["tables/inventory/interfaces", "tables/vrf/detail"],
    )
    column: Optional[str] = Field(
        None,
        description="The name of the column to use for the Attribute Value.",
        title="API Column Name",
        examples=["primaryIp", "hostname"],
    )
    sn_column: Optional[str] = Field(
        None,
        description="The name of the column to use for the device serial number, defaults to 'sn'. "
        "Some tables have a different column name for the device serial number such as 'localSn'.",
        title="Device Serial Number Column",
        examples=["sn", "localSn"],
    )
    sort: Optional[Sort] = Field(None, description="Sort the results by the specified column.", title="Table Sorting")

    @model_validator(mode="after")
    def _validate(self):
        if not self.column and not self.static:
            raise ValueError("Either 'column' or 'static' must be provided.")
        if self.column and self.static and not self.regex:
            raise ValueError("Only 'static' can only be used with 'column' and 'regex' or by itself.")
        if self.default_mapping_value and not self.mapping:
            raise ValueError("'default_mapping_value' must be used with 'mapping'.")
        if self.mapping and self.static:
            raise ValueError("Mapping cannot be used with a Static Value.")
        if self.null_value and self.static and not self.column:
            raise ValueError("'null_value' must be used with 'column'.")
        if self.regex and not self.column:
            raise ValueError("'regex' can only be used with a column.")
        return self

    @cached_property
    def sn(self) -> str:
        """Return the serial number column name to use for the API request."""
        return self.sn_column or "sn"

    def calc_value(self, data) -> tuple[Union[str, None], Union[str, None], Union[dict[str, str], None]]:
        """Calculate the value for the attribute based on the data from the API response.

        Args:
            data (dict): The data from the API response.

        Returns:
            tuple: A tuple containing the value to assign to the attribute, the regex match, and the mapping match.
        """
        if self.static and not self.column:
            # If the static value is set, we want to return it directly.
            return str(self.static), None, None
        value = data.get(self.column)
        if isinstance(value, dict) and ["data", "severity"] == list(value.keys()):
            # If the value is a dict with keys "data" and "severity", we want to extract the "data" part.
            value = value["data"]
        if value is None:
            # If the value is None, we want to return the null_value if set.
            return self.null_value, None, None

        value = str(value)
        if self.transform:
            # If a transform is set, we want to apply it to the value.
            value = value.upper() if self.transform == "upper" else value.lower()
        if self.regex:
            match = self.regex.compiled.match(value)
            if match and self.static:
                # If the static value is set, we want to return it directly on the matched value.
                return str(self.static), match.group(0), None
            if match and self.regex.group <= len(match.groups()):
                # If the regex pattern matches, we want to extract the group specified and set it as the value.
                value = match.group(self.regex.group)
            else:
                # If the regex pattern does not match, we want to return the null_value if set.
                return self.null_value, None, None
        return self.calc_mapped_value(value)



from collections import defaultdict
from pathlib import Path
from typing import Any, Union

from ipfabric import IPFClient
from ipfabric.settings.attributes import Attributes
from pandas import DataFrame
from pydantic import BaseModel, PrivateAttr, FilePath, Field

from ipfabric.dynamic_attributes.configs import Config, TableRule, ConfigRule, MultiConfig, InventoryRule, load_config
from ipfabric.dynamic_attributes.models import Attribute, MatchedRule, MatchedConfig, CONFIG_MAP


class AttributeSync(BaseModel):
    config: Union[MultiConfig, Config, FilePath, list[Union[Config, FilePath]]] = Field(
        title="Configuration File(s)",
        description="Path to the configuration file, Config object, or an ordered list of files or objects. "
        "Supports YAML, JSON, and TOML formats.",
    )
    _client: IPFClient = PrivateAttr(None)
    _attributes: dict[str, Attribute] = PrivateAttr(default_factory=dict)
    _endpoints_map: dict[str, str] = PrivateAttr(default_factory=dict)
    _endpoints: dict[str, set[str]] = PrivateAttr(default_factory=dict)
    _inventory: dict[int, set[str]] = PrivateAttr(default_factory=dict)

    def model_post_init(self, __context: Any) -> None:
        # If a single Config object is provided in a list, use it directly
        self.config = self.config[0] if isinstance(self.config, list) and len(self.config) == 1 else self.config
        if isinstance(self.config, MultiConfig):
            self.config = self.config.merge_configs()
        elif isinstance(self.config, list):
            # If a list of configs is provided, merge them into a single config
            multi = MultiConfig(configs=[_ if isinstance(_, Config) else load_config(_) for _ in self.config])
            self.config = multi.merge_configs()
        elif isinstance(self.config, Path):
            self.config = load_config(self.config)
        self._setup()

    def _validate(self, endpoints: dict[str, set[str]], dev_filters: Union[None, dict[str, set[str]]]) -> None:
        for url, columns in endpoints.items():
            if url not in self._client.oas:
                raise ValueError(f"API endpoint {url} not found in IP Fabric API.")
            if invalid := columns - set(self._client.oas[url].post.columns):
                raise ValueError(f"API endpoint '{url}' does not contain column(s): {invalid}.")
            if (
                dev_filters
                and hasattr(self._client.oas[url].post, "device_filters")
                and (invalid := dev_filters.get(url, set()) - set(self._client.oas[url].post.device_filters))
            ):
                # Added in SDK 7.2
                raise ValueError(f"API endpoint '{url}' does not contain device filter(s): {invalid}.")

    def _calculate_inventory(self, rule: Union[TableRule, ConfigRule, InventoryRule]) -> set[str]:
        # Get the filtered list of Inventory Devices for attribute calculation
        if rule.filter_hash in self._inventory:
            return self._inventory[rule.filter_hash]
        self._inventory[rule.filter_hash] = {
            _["sn"] for _ in self._client.inventory.devices.all(filters=rule.calc_filter, columns=["sn"])
        }
        return self._inventory[rule.filter_hash]

    def _setup(self) -> None:
        self._client = IPFClient(**self.config.ipfabric.model_dump())
        self._endpoints, dev_filters = defaultdict(set), defaultdict(set)

        # Validate the inventory filters
        self._validate(
            {
                "tables/inventory/devices": {
                    *{_.column for _ in self.config.inventory.filters},
                    *self.config.inventory.calc_columns,
                }
            },
            None,
        )

        # Validate the endpoints and device filters
        for rule in self.config.merged_rules:
            if isinstance(rule, ConfigRule):
                continue
            # Get all API endpoints from the rules and the columns required for each endpoint
            url = self._client._check_url(rule.value.api_endpoint)
            self._endpoints_map[rule.value.api_endpoint] = url
            endpoint = self._endpoints[url]
            endpoint.add(rule.value.sn)  # Add the serial number column
            if rule.value.column:
                endpoint.add(rule.value.column)
            if rule.value.sort:
                endpoint.add(rule.value.sort.column)
            for _ in rule.filters:
                endpoint.add(_.column)
            for _ in rule.device_filters:
                # If the filter is a device filter, add it to the dev_filters
                dev_filters[url].add(f"device.{_.column}")
        self._validate(self._endpoints, dev_filters)

        # Verify Attribute Names are formatted correctly
        attribute_names = {r.attribute for r in self.config.merged_rules}
        self._client.settings.global_attributes.check_attribute_name(attribute_names)

        # Get the filtered list of Inventory Devices for attribute calculation
        inventory = self._calculate_inventory(self.config.inventory)
        self._attributes = {_.sn: Attribute(device=_) for _ in self._client.devices.all if _.sn in inventory}

    def run(self) -> DataFrame:
        """Run the attribute sync process and return the report DataFrame."""
        self.calculate_attribute()
        report = self.report()
        return report if self.config.dry_run else self.apply_attributes(report)

    def _delete_attributes(self, report: DataFrame) -> list[str]:
        # Fetch only relevant attributes for deletion to reduce memory usage
        report_index = set(report.set_index(["sn", "attribute"]).index)
        attributes = self._client.fetch_all(
            url="tables/global-attributes",
            export="df",
            columns=["id", "sn", "name", "value"],
            snapshot=False,
        )
        attributes["delete"] = attributes.set_index(["sn", "name"]).index.isin(report_index)
        # Use list comprehension directly on the DataFrame for better performance
        if payload := attributes.loc[attributes["delete"], "id"].tolist():
            self._client.settings.global_attributes.delete_attribute_by_id(*payload)
        return payload

    def apply_attributes(self, report) -> DataFrame:
        result = {"applied": [], "deleted": []}
        payload = (
            report.query("update == True | create == True")[["sn", "attribute", "new_value"]]
            .rename(columns={"attribute": "name", "new_value": "value"})
            .to_dict("records")
        )
        if payload:
            # Submit the payload to the API
            result["applied"] = self._client.settings.global_attributes.set_attributes_by_sn(payload)

        if not (deletion := report[["sn", "attribute", "delete"]].query("delete == True")).empty:
            result["deleted"] = self._delete_attributes(deletion)

        if self.config.update_snapshot:
            # Update the snapshot with the new attributes
            Attributes(client=self._client, snapshot_id=self._client.snapshot_id).update_local_attr_from_global(
                wait_for_load=False
            )
        return report

    def _process_deletions(self, rule: Union[TableRule, ConfigRule]) -> None:
        """Process deletions for the given rule on the remaining inventory if no match was found."""
        for dev in self._attributes.values():
            current_value = dev.global_attribute(rule.attribute)
            matched_rule = dev.matched_rule.get(rule.attribute)
            if current_value and (not matched_rule or matched_rule.marked_for_deletion):
                # If a match is found, then mark for deletion
                dev.matched_rule[rule.attribute] = MatchedRule(
                    name=rule.name,
                    attribute=rule.attribute,
                    current_value=current_value,
                    delete=True,
                    overwrite=True,
                    api_endpoint=self._endpoints_map[rule.value.api_endpoint],
                    **rule.value.model_dump(exclude={"api_endpoint"}),
                )

    def _filter_device_inventory(self, rule: Union[TableRule, ConfigRule]) -> dict[str, Attribute]:
        """Filter the device inventory based on the rule's filters."""
        devices = self._attributes
        if rule.inventory and rule.inventory.calc_filter:
            # Filter inventory if multiple configurations are merged
            filtered = self._calculate_inventory(rule.inventory)
            devices = {k: v for k, v in devices.items() if k in filtered}
        if isinstance(rule, ConfigRule) and rule.calc_filter:
            filtered = self._calculate_inventory(rule)
            devices = {k: v for k, v in devices.items() if k in filtered}
        [
            dev.matched_rule.update({rule.attribute: None})
            for dev in devices.values()
            if rule.attribute not in dev.matched_rule
        ]
        return devices

    def _process_table_rule(self, rule: TableRule) -> None:
        """Process Table Rules if a match/new_value has not been applied."""

        # Filter the device inventory based on the rule's filters
        inventory = self._filter_device_inventory(rule)

        url = self._endpoints_map[rule.value.api_endpoint]
        for data in self._client.fetch_all(
            url,
            filters=rule.calc_filter,
            columns=list(self._endpoints[url]),
            sort=rule.value.sort.model_dump() if rule.value.sort else None,
        ):
            dev = inventory.get(data[rule.value.sn])
            if not dev:
                continue
            matched_rule = dev.matched_rule.get(rule.attribute)
            if matched_rule and matched_rule.marked_for_deletion is False:
                continue
            value, regex_match, map_dict = rule.value.calc_value(data)
            if value:
                # If a match is found, assign the value to the device's attribute
                dev.matched_rule[rule.attribute] = MatchedRule(
                    name=rule.name,
                    attribute=rule.attribute,
                    delete=False,
                    overwrite=bool(rule.overwrite),
                    api_endpoint=url,
                    current_value=dev.global_attribute(rule.attribute),
                    new_value=value,
                    regex_match=regex_match,
                    column_value=data.get(rule.value.column) if rule.value.column else None,
                    mapping_match=map_dict,
                    **rule.value.model_dump(exclude={"api_endpoint"}),
                )

    @staticmethod
    def _get_config_or_log(dev: Attribute, rule: ConfigRule, match: MatchedConfig) -> Union[None, str]:
        if rule.value.config == "log":
            config = dev.device.get_log_file()
            if not config:
                match.new_value = rule.value.no_log_value
                dev.no_log_file = True
                dev.matched_rule[rule.attribute] = match
                return None
            setattr(dev, CONFIG_MAP[rule.value.config]["config"], config)
            return config

        config = dev.device.get_config()
        if not config:
            if (rule.overwrite_unsupported and match.current_value) or not match.current_value:
                match.new_value = rule.value.no_config_value
            else:
                match.new_value = match.current_value
            dev.no_start_config, dev.no_current_config = True, True
            dev.matched_rule[rule.attribute] = match
            return None

        config_string = getattr(config, CONFIG_MAP[rule.value.config]["name"], None)
        if not config_string:
            match.new_value = rule.value.no_config_value
            setattr(dev, CONFIG_MAP[rule.value.config]["no"], True)
            dev.matched_rule[rule.attribute] = match
            return None
        setattr(dev, CONFIG_MAP[rule.value.config]["config"], config_string)
        return config_string

    def _get_config(self, dev: Attribute, rule: ConfigRule) -> Union[None, str]:
        if config_string := getattr(dev, CONFIG_MAP[rule.value.config]["config"], None):
            return config_string

        match = MatchedConfig(
            name=rule.name,
            attribute=rule.attribute,
            current_value=dev.global_attribute(rule.attribute),
            delete=False,
            no_config_match=True,
            **rule.value.model_dump(),
        )
        return self._get_config_or_log(dev, rule, match)

    def _process_config_rule(self, rule: ConfigRule) -> None:
        """Process config rules if a match/new_value has not been applied."""

        # Filter the device inventory based on the rule's filters
        inventory = self._filter_device_inventory(rule)

        for dev in inventory.values():
            if dev.config_skip(rule) or (not (config_string := self._get_config(dev, rule))):
                continue
            new_value, regex_match, mapping_match = rule.value.calc_value(config_string)
            if new_value:
                dev.matched_rule[rule.attribute] = MatchedConfig(
                    name=rule.name,
                    attribute=rule.attribute,
                    current_value=dev.global_attribute(rule.attribute),
                    delete=False,
                    overwrite=bool(rule.overwrite),
                    no_config_match=False,
                    new_value=new_value,
                    regex_match=regex_match,
                    mapping_match=mapping_match,
                    **rule.value.model_dump(),
                )

    def calculate_attribute(self) -> dict[str, Attribute]:
        for rule in self.config.merged_rules:
            if isinstance(rule, TableRule):
                self._process_table_rule(rule)
            else:
                self._process_config_rule(rule)
            if rule.delete_attribute:
                self._process_deletions(rule)
        return self._attributes

    def report(self) -> DataFrame:
        df = DataFrame(
            [_ for dev in self._attributes.values() for _ in dev.export(self.config.inventory.calc_columns)],
            columns=self.config.inventory.df_columns,
        )
        df["correct"] = df["new_value"] == df["global_value"]
        df["update"] = df["new_value"].notnull() & (df["new_value"] != df["global_value"]) & df["overwrite"]
        df["create"] = df["new_value"].notnull() & df["global_value"].isnull()
        return df



from typing import Optional, Literal, Annotated, Union
from uuid import uuid4

from pydantic import BaseModel, Field


class BaseRule(BaseModel):
    rule_type: Literal["regexHostname", "regexSnmpLocation", "regexCloudResourceId", "slug", "routingSwitching"] = (
        Field(alias="type")
    )
    rule_id: str = Field(alias="id", default_factory=lambda x: str(x or uuid4()))
    note: str = ""
    transformation: Literal["none", "uppercase", "lowercase"]


class SiteRule(BaseRule, BaseModel):
    siteName: Optional[str] = None
    regex: str


class HostnameRule(SiteRule, BaseModel):
    rule_type: Literal["regexHostname"] = Field(alias="type")


class SNMPRule(SiteRule, BaseModel):
    rule_type: Literal["regexSnmpLocation"] = Field(alias="type")


class RouteSwitchRule(BaseRule, BaseModel):
    rule_type: Literal["routingSwitching"] = Field(alias="type")
    firewall: bool
    minimumSiteSize: int = Field(ge=1)
    wan: list[Literal["serial", "tunnel"]] = Field(default_factory=list)


class SlugRule(BaseRule, BaseModel):
    rule_type: Literal["slug"] = Field(alias="type")
    applyToCloudInstances: bool


class CloudIDRule(SiteRule, BaseModel):
    rule_type: Literal["regexCloudResourceId"] = Field(alias="type")
    applyToCloudInstances: bool


SeparationRule = Annotated[
    Union[
        HostnameRule,
        SNMPRule,
        RouteSwitchRule,
        SlugRule,
        CloudIDRule,
    ],
    Field(discriminator="rule_type"),
]


class SiteSeparation(BaseModel):
    rules: list[SeparationRule] = Field(default_factory=list)
    manualEnabled: bool
    neighborshipFallbackEnabled: bool



import logging
from typing import Optional, Any, Union, overload, Literal

from pydantic import BaseModel, computed_field, Field

from .device import Devices
from .table import Table

try:
    from pandas import DataFrame
except ImportError:
    DataFrame = None
logger = logging.getLogger("ipfabric")

IGNORE_COLUMNS = {"id"}
EXPORT_FORMAT = Literal["json", "csv", "df"]
DEVICE_EXPORT_FORMAT = Literal["json", "csv", "object", "df"]


class DeviceTable(Table):
    """model for Device Table data"""

    endpoint: str = "tables/inventory/devices"

    def _as_model(self, devices, export):
        if export != "object":
            return devices
        return Devices(
            client=self.client,
            snapshot_id=self.client.snapshot_id,
            devices=devices,
        )

    @overload
    def fetch(
        self,
        export: Literal["json"] = ...,
        columns: Optional[Union[list[str], set[str]]] = None,
        filters: Optional[dict] = None,
        attr_filters: Optional[dict[str, list[str]]] = None,
        snapshot_id: Optional[str] = None,
        reports: Optional[Union[bool, list, str]] = False,
        sort: Optional[dict] = None,
        limit: Optional[int] = 1000,
        start: Optional[int] = 0,
        api_version: Optional[Union[str, int]] = None,
    ) -> list[dict]: ...

    @overload
    def fetch(
        self,
        export: Literal["csv"],
        columns: Optional[Union[list[str], set[str]]] = None,
        filters: Optional[dict] = None,
        attr_filters: Optional[dict[str, list[str]]] = None,
        snapshot_id: Optional[str] = None,
        sort: Optional[dict] = None,
        limit: Optional[int] = 1000,
        start: Optional[int] = 0,
        csv_tz: Optional[str] = None,
        # api_version: Optional[Union[str, int]] = None,  # TODO: NIM-21720
    ) -> bytes: ...

    @overload
    def fetch(
        self,
        export: Literal["df"],
        columns: Optional[Union[list[str], set[str]]] = None,
        filters: Optional[dict] = None,
        attr_filters: Optional[dict[str, list[str]]] = None,
        snapshot_id: Optional[str] = None,
        reports: Optional[Union[bool, list, str]] = False,
        sort: Optional[dict] = None,
        limit: Optional[int] = 1000,
        start: Optional[int] = 0,
        api_version: Optional[Union[str, int]] = None,
    ) -> DataFrame: ...

    @overload
    def fetch(
        self,
        export: Literal["object"],
        columns: Optional[Union[list[str], set[str]]] = None,
        filters: Optional[dict] = None,
        attr_filters: Optional[dict[str, list[str]]] = None,
        snapshot_id: Optional[str] = None,
        sort: Optional[dict] = None,
        limit: Optional[int] = 1000,
        start: Optional[int] = 0,
    ) -> Devices: ...

    def fetch(
        self,
        export: DEVICE_EXPORT_FORMAT = "json",
        columns: Optional[Union[list[str], set[str]]] = None,
        filters: Optional[dict] = None,
        attr_filters: Optional[dict[str, list[str]]] = None,
        snapshot_id: Optional[str] = None,
        reports: Optional[Union[bool, list, str]] = False,
        sort: Optional[dict] = None,
        limit: Optional[int] = 1000,
        start: Optional[int] = 0,
        csv_tz: Optional[str] = None,
        api_version: Optional[Union[str, int]] = None,
    ):
        """Gets all data from corresponding endpoint

        Args:
            export: str: Export format to return [json, csv, object]; default is json.
            columns: Optional columns to return, default is all
            filters: Optional filters'
            attr_filters: dict: Optional dictionary of Attribute filters
            snapshot_id: Optional snapshot ID to override class
            reports: True to return Intent Rules (also accepts string of frontend URL) or a list of report IDs
            sort: Dictionary to apply sorting: {"order": "desc", "column": "lastChange"}
            limit: Default to 1,000 rows
            start: Starts at 0
            csv_tz: str: Default None, set a timezone to return human-readable dates when using CSV;
                         see `ipfabric.tools.shared.TIMEZONES`
            api_version: Optional API version to use for this request's X-API-Version header,
                         default None will use latest version. Values other than None will not use streaming requests
                         and will switch to pagination. API Version is not supported with CSV export.
        Returns:
            Union[list[dict], Devices, bytes, DataFrame]: Default List of dicts 'json', Devices object if 'object',
                                                          bytes if 'csv', pandas.DataFrame if 'df'
        """
        devices = super(DeviceTable, self).fetch(
            export="json" if export in ["json", "object"] else export,
            columns=columns,
            filters=filters,
            attr_filters=attr_filters,
            snapshot_id=snapshot_id,
            reports=reports if export not in ["csv", "object"] else None,
            sort=sort,
            limit=limit,
            start=start,
            csv_tz=csv_tz,
            api_version=api_version if export != "object" else None,
        )
        return self._as_model(devices, export)

    @overload
    def all(
        self,
        export: Literal["json"] = ...,
        columns: Optional[Union[list[str], set[str]]] = None,
        filters: Optional[dict] = None,
        attr_filters: Optional[dict[str, list[str]]] = None,
        snapshot_id: Optional[str] = None,
        reports: Optional[Union[bool, list, str]] = False,
        sort: Optional[dict] = None,
        api_version: Optional[Union[str, int]] = None,
    ) -> list[dict]: ...

    @overload
    def all(
        self,
        export: Literal["csv"],
        columns: Optional[Union[list[str], set[str]]] = None,
        filters: Optional[dict] = None,
        attr_filters: Optional[dict[str, list[str]]] = None,
        snapshot_id: Optional[str] = None,
        sort: Optional[dict] = None,
        csv_tz: Optional[str] = None,
        # api_version: Optional[Union[str, int]] = None,  # TODO: NIM-21720
    ) -> bytes: ...

    @overload
    def all(
        self,
        export: Literal["df"],
        columns: Optional[Union[list[str], set[str]]] = None,
        filters: Optional[dict] = None,
        attr_filters: Optional[dict[str, list[str]]] = None,
        snapshot_id: Optional[str] = None,
        reports: Optional[Union[bool, list, str]] = False,
        sort: Optional[dict] = None,
        api_version: Optional[Union[str, int]] = None,
    ) -> DataFrame: ...

    @overload
    def all(
        self,
        export: Literal["object"],
        columns: Optional[Union[list[str], set[str]]] = None,
        filters: Optional[dict] = None,
        attr_filters: Optional[dict[str, list[str]]] = None,
        snapshot_id: Optional[str] = None,
        sort: Optional[dict] = None,
    ) -> Devices: ...

    def all(
        self,
        export: DEVICE_EXPORT_FORMAT = "json",
        columns: Optional[Union[list[str], set[str]]] = None,
        filters: Optional[dict] = None,
        attr_filters: Optional[dict[str, list[str]]] = None,
        snapshot_id: Optional[str] = None,
        reports: Optional[Union[bool, list, str]] = False,
        sort: Optional[dict] = None,
        csv_tz: Optional[str] = None,
        api_version: Optional[Union[str, int]] = None,
    ):
        """Gets all data from corresponding endpoint

        Args:
            export: str: Export format to return [json, csv, object]; default is json.
            columns: Optional columns to return, default is all
            filters: Optional filters
            attr_filters: dict: Optional dictionary of Attribute filters
            snapshot_id: Optional snapshot ID to override class
            reports: True to return Intent Rules (also accepts string of frontend URL) or a list of report IDs
            sort: Dictionary to apply sorting: {"order": "desc", "column": "lastChange"}
            csv_tz: str: Default None, set a timezone to return human-readable dates when using CSV;
                         see `ipfabric.tools.shared.TIMEZONES`
            api_version: Optional API version to use for this request's X-API-Version header,
                         default None will use latest version. Values other than None will not use streaming requests
                         and will switch to pagination. API Version is not supported with CSV export.
        Returns:
             Union[list[dict], Devices, bytes, DataFrame]: Default List of dicts 'json', Devices object if 'object',
                                                           bytes if 'csv', pandas.DataFrame if 'df'
        """
        devices = super(DeviceTable, self).all(
            export="json" if export in ["json", "object"] else export,
            columns=columns,
            filters=filters,
            attr_filters=attr_filters,
            snapshot_id=snapshot_id,
            reports=reports if export not in ["csv", "object"] else None,
            sort=sort,
            csv_tz=csv_tz,
            api_version=api_version if export != "object" else None,
        )
        return self._as_model(devices, export)


class Inventory(BaseModel):
    """model for inventories"""

    client: Any = Field(None, exclude=True)

    def print_tables(self):
        print(sorted([_ for _ in dir(self) if _[0] != "_" and isinstance(getattr(self, _), Table)]))

    @computed_field
    @property
    def sites(self) -> Table:
        return Table(client=self.client, endpoint="tables/inventory/sites")

    @computed_field
    @property
    def vendors(self) -> Table:
        return Table(client=self.client, endpoint="tables/inventory/summary/vendors")

    @computed_field
    @property
    def devices(self) -> DeviceTable:
        return DeviceTable(client=self.client)

    @computed_field
    @property
    def models(self) -> Table:
        return Table(client=self.client, endpoint="tables/inventory/summary/models")

    @computed_field
    @property
    def os_version_consistency_models(self) -> Table:
        return Table(client=self.client, endpoint="tables/inventory/os-version-consistency/models")

    @computed_field
    @property
    def os_version_consistency_platforms(self) -> Table:
        return Table(client=self.client, endpoint="tables/inventory/os-version-consistency/platforms")

    @computed_field
    @property
    def eol_summary(self) -> Table:
        return Table(client=self.client, endpoint="tables/reports/eof/summary")

    @computed_field
    @property
    def eol_details(self) -> Table:
        return Table(client=self.client, endpoint="tables/reports/eof/detail")

    @computed_field
    @property
    def platforms(self) -> Table:
        return Table(client=self.client, endpoint="tables/inventory/summary/platforms")

    @computed_field
    @property
    def pn(self) -> Table:
        return Table(client=self.client, endpoint="tables/inventory/pn")

    @computed_field
    @property
    def families(self) -> Table:
        return Table(client=self.client, endpoint="tables/inventory/summary/families")

    @computed_field
    @property
    def interfaces(self) -> Table:
        return Table(client=self.client, endpoint="tables/inventory/interfaces")

    @computed_field
    @property
    def hosts(self) -> Table:
        return Table(client=self.client, endpoint="tables/addressing/hosts")

    @computed_field
    @property
    def hosts_ipv6(self) -> Table:
        return Table(client=self.client, endpoint="tables/addressing/ipv6-hosts")

    @computed_field
    @property
    def phones(self) -> Table:
        return Table(client=self.client, endpoint="tables/inventory/phones")

    @computed_field
    @property
    def fans(self) -> Table:
        return Table(client=self.client, endpoint="tables/inventory/fans")

    @computed_field
    @property
    def modules(self) -> Table:
        return Table(client=self.client, endpoint="tables/inventory/modules")



import logging
from typing import Any, Optional

from pydantic import BaseModel, Field, computed_field

from ipfabric.models.table import Table

logger = logging.getLogger("ipfabric")


class Cloud(BaseModel):
    client: Any = Field(None, exclude=True)
    sn: Optional[str] = None

    def print_tables(self):
        print(sorted([_ for _ in dir(self) if _[0] != "_" and isinstance(getattr(self, _), Table)]))

    @computed_field
    @property
    def virtual_machines(self) -> Table:
        return Table(client=self.client, endpoint="tables/cloud/endpoints/virtual-machines", sn=self.sn)

    @computed_field
    @property
    def virtual_machines_interfaces(self) -> Table:
        return Table(client=self.client, endpoint="tables/cloud/endpoints/virtual-machines-interfaces", sn=self.sn)

    @computed_field
    @property
    def endpoint_groups(self) -> Table:
        return Table(client=self.client, endpoint="tables/cloud/endpoints/endpoint-groups", sn=self.sn)

    @computed_field
    @property
    def private_link_endpoints(self) -> Table:
        return Table(client=self.client, endpoint="tables/cloud/private-link/endpoints", sn=self.sn)

    @computed_field
    @property
    def private_link_services_connections(self) -> Table:
        return Table(client=self.client, endpoint="tables/cloud/private-link/services/connections", sn=self.sn)

    @computed_field
    @property
    def private_link_services_inventory(self) -> Table:
        return Table(client=self.client, endpoint="tables/cloud/private-link/services/inventory", sn=self.sn)

    @computed_field
    @property
    def public_ips_ipv4_addresses(self) -> Table:
        return Table(client=self.client, endpoint="tables/cloud/public-ips/ipv4-addresses", sn=self.sn)

    @computed_field
    @property
    def public_ips_ipv6_addresses(self) -> Table:
        return Table(client=self.client, endpoint="tables/cloud/public-ips/ipv6-addresses", sn=self.sn)

    @computed_field
    @property
    def subnets_inventory(self) -> Table:
        return Table(client=self.client, endpoint="tables/cloud/subnets/inventory", sn=self.sn)

    @computed_field
    @property
    def aws_inventory(self) -> Table:
        return Table(client=self.client, endpoint="tables/cloud/vendors/aws/inventory", sn=self.sn)

    @computed_field
    @property
    def aws_private_link_vpc_endpoints(self) -> Table:
        return Table(client=self.client, endpoint="tables/cloud/vendors/aws/private-link/vpc-endpoints", sn=self.sn)

    @computed_field
    @property
    def aws_scaling_groups(self) -> Table:
        return Table(client=self.client, endpoint="tables/cloud/vendors/aws/scaling-groups", sn=self.sn)

    @computed_field
    @property
    def aws_subnets(self) -> Table:
        return Table(client=self.client, endpoint="tables/cloud/vendors/aws/subnets", sn=self.sn)

    @computed_field
    @property
    def azure_inventory(self) -> Table:
        return Table(client=self.client, endpoint="tables/cloud/vendors/azure/inventory", sn=self.sn)

    @computed_field
    @property
    def azure_private_link_connections(self) -> Table:
        return Table(client=self.client, endpoint="tables/cloud/vendors/azure/private-link/connections", sn=self.sn)

    @computed_field
    @property
    def azure_private_link_endpoints(self) -> Table:
        return Table(client=self.client, endpoint="tables/cloud/vendors/azure/private-link/endpoints", sn=self.sn)

    @computed_field
    @property
    def azure_private_link_services(self) -> Table:
        return Table(client=self.client, endpoint="tables/cloud/vendors/azure/private-link/services", sn=self.sn)

    @computed_field
    @property
    def azure_service_endpoints(self) -> Table:
        return Table(client=self.client, endpoint="tables/cloud/vendors/azure/service-endpoints", sn=self.sn)

    @computed_field
    @property
    def azure_subnets(self) -> Table:
        return Table(client=self.client, endpoint="tables/cloud/vendors/azure/subnets", sn=self.sn)

    @computed_field
    @property
    def azure_virtual_machine_scale_sets(self) -> Table:
        return Table(client=self.client, endpoint="tables/cloud/vendors/azure/virtual-machine-scale-sets", sn=self.sn)

    @computed_field
    @property
    def azure_vnet_integration_app_services(self) -> Table:
        return Table(
            client=self.client, endpoint="tables/cloud/vendors/azure/vnet-integration/app-services", sn=self.sn
        )

    @computed_field
    @property
    def azure_vnet_integration_container_apps(self) -> Table:
        return Table(
            client=self.client, endpoint="tables/cloud/vendors/azure/vnet-integration/container-apps", sn=self.sn
        )

    @computed_field
    @property
    def azure_vnet_integration_flexible_servers(self) -> Table:
        return Table(
            client=self.client, endpoint="tables/cloud/vendors/azure/vnet-integration/flexible-servers", sn=self.sn
        )

    @computed_field
    @property
    def azure_vnet_integration_sql_managed_instances(self) -> Table:
        return Table(
            client=self.client, endpoint="tables/cloud/vendors/azure/vnet-integration/sql-managed-instances", sn=self.sn
        )

    @computed_field
    @property
    def gcp_inventory(self) -> Table:
        return Table(client=self.client, endpoint="tables/cloud/vendors/gcp/inventory", sn=self.sn)

    @computed_field
    @property
    def gcp_instance_network_endpoint_groups(self) -> Table:
        return Table(
            client=self.client, endpoint="tables/cloud/vendors/gcp/instance-network-endpoint-groups", sn=self.sn
        )

    @computed_field
    @property
    def gcp_private_service_connect_endpoints(self) -> Table:
        return Table(
            client=self.client, endpoint="tables/cloud/vendors/gcp/private-service-connect/endpoints", sn=self.sn
        )

    @computed_field
    @property
    def gcp_private_service_connect_published_services(self) -> Table:
        return Table(
            client=self.client,
            endpoint="tables/cloud/vendors/gcp/private-service-connect/published-services",
            sn=self.sn,
        )

    @computed_field
    @property
    def gcp_private_service_connect_published_services_connections(self) -> Table:
        return Table(
            client=self.client,
            endpoint="tables/cloud/vendors/gcp/private-service-connect/published-services-connections",
            sn=self.sn,
        )

    @computed_field
    @property
    def gcp_subnets(self) -> Table:
        return Table(client=self.client, endpoint="tables/cloud/vendors/gcp/subnets", sn=self.sn)

    @computed_field
    @property
    def inventory(self) -> Table:
        return Table(client=self.client, endpoint="tables/cloud/nodes/inventory", sn=self.sn)

    @computed_field
    @property
    def tags(self) -> Table:
        return Table(client=self.client, endpoint="tables/cloud/nodes/tags", sn=self.sn)



import argparse
import json
import os
import re
from pathlib import Path

import niquests

from ipfabric import IPFClient
from ipfabric.models import Inventory, Technology, Endpoint, Device
from ipfabric.scripts.shared import base_args, load_env, parse_base_args

try:
    import pandas as pd
    import openpyxl  # noqa: F401
    from rich.console import Console
    from rich import box  # Export BOX_STYLE=MARKDOWN before running commands for docs.
    from rich.table import Table
except ImportError:
    raise ImportError(
        "pandas, openpyxl, and rich are required, please install by using: "
        "`pip3 install ipfabric[cli]` or `pip3 install pandas openpyxl rich`."
    )

CONSOLE = Console()


def collect_endpoints(data):
    endpoints = set()
    for key, value in data.items():
        if isinstance(value, dict):
            if "endpoint" in value:
                endpoints.add(value["endpoint"])
            endpoints.update(collect_endpoints(value))
    return endpoints


def verify_tables(ipf: IPFClient, args: argparse.Namespace) -> set[Endpoint]:
    valid_tables = set()
    for table in args.ipf_tables:
        if not (endpoint := ipf.oas.get(ipf._check_url(table), None)) or not endpoint.post:
            CONSOLE.print(f"[red]Error:[/red] Not a valid table: {table}")
            continue
        if (args.device_hostname or args.device_sn) and not endpoint.post.sn_columns:
            CONSOLE.print(f"[red]Error:[/red] Table '{table}' does not have 'sn' column.")
            continue
        valid_tables.add(endpoint.post)
    return valid_tables


def rich_print(results: pd.DataFrame, filters: dict, url: str, columns: list[str]):

    # Export BOX_STYLE=MARKDOWN before running commands for docs.
    box_style = getattr(box, os.getenv("BOX_STYLE", "HEAVY_HEAD").upper(), box.HEAVY_HEAD)
    caption = f"Filter: {json.dumps(filters)}" if filters else None
    table = Table(*columns, title=results.title, caption=caption, box=box_style)

    for result in results.head(3).to_dict(orient="records"):
        table.add_row(*[str(result[c]) for c in columns])
    CONSOLE.print(table)
    CONSOLE.print(url, style=f"link {url}")


def get_table(
    ipf: IPFClient, args: argparse.Namespace, table: Endpoint, table_filter: dict, title: str
) -> pd.DataFrame | None:
    try:
        data = ipf.fetch_all(table.api_endpoint, filters=table_filter, export="df")
        data.title = title
        if args.preview and data is not None:
            url = ipf.oas[table.api_endpoint].post.filter_url(table_filter, ipf._client.base_url)
            rich_print(data, table_filter, url, table.columns)
        return data
    except niquests.exceptions.HTTPError as e:
        if e.response.status_code == 422:
            CONSOLE.print(f"Cant fetch data for {table}.")
        CONSOLE.print(f"[red]Error:[/red] {e} ")


def create_filter(devices: list[Device], table: Endpoint) -> dict:
    return {"or": [{col: ["eq", dev.sn]} for dev in devices for col in table.sn_columns]} if devices else {}


def get_tables_by_device(
    ipf: IPFClient, args: argparse.Namespace, valid_tables: set[Endpoint], valid_devices: set[Device]
) -> list[pd.DataFrame]:
    dataframes_for_export = []

    for table in valid_tables:
        for device in valid_devices:
            title = f"{device.hostname}-{table.title or table.summary}"
            if (data := get_table(ipf, args, table, create_filter([device], table), title)) is not None:
                dataframes_for_export.append(data)
    return dataframes_for_export


def get_tables_joined(
    ipf: IPFClient, args: argparse.Namespace, valid_tables: set[Endpoint], valid_devices: set[Device]
) -> list[pd.DataFrame]:
    dataframes_for_export = []

    for table in valid_tables:
        table_filter = create_filter(valid_devices, table)
        title = table.title or table.summary
        if (data := get_table(ipf, args, table, table_filter, title)) is not None:
            dataframes_for_export.append(data)
    return dataframes_for_export


def get_tables(args: argparse.Namespace) -> list[pd.DataFrame]:
    ipf = IPFClient(snapshot_id=args.snapshot, base_url=args.base_url, auth=args.auth, verify=(not args.insecure))
    valid_tables = verify_tables(ipf, args)
    if not valid_tables:
        CONSOLE.print("[red]Error:[/red] No valid tables provided. Exiting...")
        exit(1)

    valid_devices = set()
    for hostname in args.device_hostname:
        if hostname not in ipf.devices.by_hostname:
            CONSOLE.print(f"[red]Error:[/red] Hostname '{hostname}' not found in inventory.")
            continue
        valid_devices.add(ipf.devices.by_hostname[hostname][0])
    for sn in args.device_sn:
        if sn not in ipf.devices.by_sn:
            CONSOLE.print(f"[red]Error:[/red] Device Serial Number '{sn}' not found in inventory.")
            continue
        valid_devices.add(ipf.devices.by_sn[sn])
    if (args.device_hostname or args.device_sn) and not valid_devices:
        CONSOLE.print("[red]Error:[/red] No valid hostnames provided. Exiting...")
        exit(1)

    return (
        get_tables_by_device(ipf, args, valid_tables, valid_devices)
        if args.device_separate
        else get_tables_joined(ipf, args, valid_tables, valid_devices)
    )


def main():
    load_env()
    parser = argparse.ArgumentParser(
        description="Fetch tables from IP Fabric and insert into a single excel file with multiple sheets."
    )
    CONSOLE.print(":rocket: IP Fabric Tables to Excel Exporter! :rocket:")

    ipf_tables = collect_endpoints(Inventory(client=None).model_dump())
    ipf_tables.update(collect_endpoints(Technology(client=None).model_dump()))
    parser.add_argument(
        "ipf_tables",
        help="API or Front end URL for IP Fabric tables to fetch. Can be used multiple times.",
        metavar="table",
        nargs="+",
    )
    parser.add_argument(
        "-d",
        "--device-hostname",
        help="Device Hostname to fetch data for. Can be used multiple times; default to return all devices.",
        metavar="hostname",
        nargs="+",
        action="extend",
        default=[],
    )
    parser.add_argument(
        "-sn",
        "--device-sn",
        help="Device Serial Number to fetch data for. Can be used multiple times; default to return all devices.",
        metavar="sn",
        nargs="+",
        action="extend",
        default=[],
    )
    parser.add_argument(
        "-P",
        "--preview",
        action="store_true",
        help="Print first 3 rows of each table fetched",
        default=False,
    )
    parser.add_argument(
        "-D",
        "--device-separate",
        action="store_true",
        help="Creates a unique Excel Worksheet separately for each device for each table requested. "
        "Default is False and will return a single Worksheet for each table requested.",
        default=False,
    )
    parser.add_argument(
        "-f",
        "--filename",
        type=str,
        help="Name or Path of the output file",
        default="output.xlsx",
    )
    parser = base_args(parser)
    args = parse_base_args(parser.parse_args())
    if not args.ipf_tables:
        CONSOLE.print("[red]Error:[/red] No tables provided. Exiting...")
        exit(1)

    dataframes_for_export = get_tables(args)

    CONSOLE.print("Exporting data to excel...")

    args.filename = Path(args.filename).resolve().with_suffix(".xlsx").absolute()
    with pd.ExcelWriter(f"{args.filename}", engine="openpyxl") as writer:
        for index, df in enumerate(dataframes_for_export):
            sheet_name = re.sub(r"[\/\\\?\*\:\[\]]", "_", df.title)
            if len(sheet_name) > 31:
                CONSOLE.print(f"[yellow]Sheet name {sheet_name} is too long.[/yellow] Truncating to 31 characters.")
            df.to_excel(writer, sheet_name=f"{sheet_name[:31]}", index=False)
    CONSOLE.print(f"Export complete. Check {args.filename}")

    CONSOLE.print(":wave: Bye!")


if __name__ == "__main__":
    main()



import logging
from typing import Optional, Any, Union, ClassVar

from dateutil import parser
from pydantic import BaseModel, Field, PrivateAttr, model_serializer

from ipfabric.models.authentication import Credential, Privilege
from ipfabric.tools.shared import validate_ip_network_str, raise_for_status

logger = logging.getLogger("ipfabric")
CRED_ERROR = "Creating credentials in a snapshot is not supported."


class Authentication(BaseModel):
    client: Any = Field(exclude=True)
    snapshot_id: Optional[str] = Field(None, exclude=True)
    settings: Optional[dict] = Field(None, exclude=True)
    _credentials: dict[int, Credential] = PrivateAttr(default={})
    _privileges: dict[int, Privilege] = PrivateAttr(default={})
    _cred_url: ClassVar[str] = "settings/credentials"
    _priv_url: ClassVar[str] = "settings/privileges"

    @property
    def _settings_url(self) -> str:
        return f"snapshots/{self.snapshot_id}/settings" if self.snapshot_id else "settings"

    def model_post_init(self, __context: Any) -> None:
        if self.settings:
            self._credentials = {cred["priority"]: Credential(**cred) for cred in self.settings["credentials"]}
            self._privileges = {priv["priority"]: Privilege(**priv) for priv in self.settings["privileges"]}
        else:
            self._credentials = self.get_credentials()
            self._privileges = self.get_enables()

    @model_serializer
    def _ser_model(self) -> dict[str, Any]:
        return {
            "credentials": [_.model_dump(by_alias=True) for _ in self.credentials.values()],
            "privileges": [_.model_dump(by_alias=True) for _ in self.enables.values()],
        }

    @property
    def credentials(self) -> dict[int, Credential]:
        return self._credentials

    @property
    def enables(self) -> dict[int, Privilege]:
        return self._privileges

    @property
    def credentials_by_id(self) -> dict[str, Credential]:
        return {_.credential_id: _ for _ in self._credentials.values()}

    @property
    def enables_by_id(self) -> dict[str, Privilege]:
        return {_.privilege_id: _ for _ in self._privileges.values()}

    def get_credentials(self) -> dict[int, Credential]:
        """
        Get all credentials and sets them in the Authentication.credentials
        :return: self.credentials
        """
        self._credentials = {
            cred["priority"]: Credential(**cred)
            for cred in raise_for_status(self.client.get(self._settings_url)).json()["credentials"]
        }
        return self._credentials

    def get_enables(self) -> dict[int, Privilege]:
        """
        Get all privileges (enable passwords) and sets them in the Authentication.enables
        :return:
        """
        self._privileges = {
            priv["priority"]: Privilege(**priv)
            for priv in raise_for_status(self.client.get(self._settings_url)).json()["privileges"]
        }
        return self._privileges

    @staticmethod
    def _create_payload(username, password, notes, network, excluded, expiration):
        networks = network or ["0.0.0.0/0"]
        excluded = excluded or []
        if expiration:
            expires = {
                "enabled": True,
                "value": parser.parse(expiration).strftime("%Y-%m-%d %H:%M:%S"),
            }
        else:
            expires = {"enabled": False}
        payload = {
            "password": password,
            "username": username,
            "notes": notes or username,
            "excludeNetworks": [validate_ip_network_str(e, ipv6=True) for e in excluded],
            "expirationDate": expires,
            "network": [validate_ip_network_str(e, ipv6=True) for e in networks],
        }
        return payload

    def update_credential(
        self,
        credential: Credential,
        password: Optional[str] = None,
    ) -> Credential:
        """
        Updates a credential. Username cannot be changed, please delete and recreate.

        :param credential: Credential: Modified Credential object to update
        :param password: str: If updating the password then this is the unencrypted password

        :return: Credential: Obj: Credential Obj with ID and encrypted password
        """
        all_creds = {v.credential_id: v.model_dump(by_alias=True) for v in self.credentials.values()}
        if all_creds[credential.credential_id]["username"] != credential.username:
            raise SyntaxError("Cannot change username of credential, please delete and recreate.")
        if all_creds[credential.credential_id]["priority"] != credential.priority:
            raise SyntaxError(
                "Updating the priority is not supported by this method, please use `update_cred_priority`."
            )
        all_creds[credential.credential_id] = credential.model_dump(by_alias=True)
        if password:
            all_creds[credential.credential_id].update({"password": password, "custom": True})
        raise_for_status(self.client.patch(self._settings_url, json={"credentials": list(all_creds.values())}))
        self.get_credentials()
        cred = self.credentials_by_id[credential.credential_id]
        logger.info(f"Updated credential with username {cred.username} and ID of {cred.credential_id}")
        return cred

    def create_credential(
        self,
        username: str,
        password: str,
        networks: list = None,
        notes: str = None,
        excluded: list = None,
        config_mgmt: bool = True,
        expiration: str = None,
    ) -> Credential:
        """
        Creates a new credential. Requires username and password and will default to all networks with no expiration.
        Does not default to use for configuration management, please set to true if needed.
        After creation Authentication.credentials will be updated with new priorities and the new cred is returned.
        :param username: str: Username
        :param password: str: Unencrypted password
        :param networks: list: List of networks defaults to ["0.0.0.0/0"]
        :param notes: str: Optional Note/Description of credential
        :param excluded: list: Optional list of networks to exclude
        :param config_mgmt: bool: Default True - do not use for configuration management
        :param expiration: str: Optional date for expiration, if none then do not expire.
                                To ensure correct date use YYYYMMDD or MM/DD/YYYY formats
        :return: Credential: Obj: Credential Obj with ID and encrypted password
        """
        if self.snapshot_id:
            raise NotImplementedError(CRED_ERROR)
        payload = self._create_payload(username, password, notes, networks, excluded, expiration)
        payload.update({"syslog": config_mgmt})
        res = raise_for_status(self.client.post(self._cred_url, json=payload))
        self.get_credentials()
        cred = Credential(**res.json())
        logger.info(f"Created credential with username {cred.username} and ID of {cred.credential_id}")
        return cred

    def update_enable(
        self,
        enable: Privilege,
        password: Optional[str] = None,
    ) -> Credential:
        """
        Updates a enable privilege password. Username cannot be changed, please delete and recreate.

        :param enable: Privilege: Modified Credential object to update
        :param password: str: If updating the password then this is the unencrypted password

        :return: Privilege: Obj: Privilege Obj with ID and encrypted password
        """
        all_priv = {v.privilege_id: v.model_dump(by_alias=True) for v in self.enables.values()}
        if all_priv[enable.privilege_id]["username"] != enable.username:
            raise SyntaxError("Cannot change username of enable privilege, please delete and recreate.")
        if all_priv[enable.privilege_id]["priority"] != enable.priority:
            raise SyntaxError(
                "Updating the priority is not supported by this method, please use `update_enable_priority`."
            )
        all_priv[enable.privilege_id] = enable.model_dump(by_alias=True)
        if password:
            all_priv[enable.privilege_id].update({"password": password, "custom": True})
        raise_for_status(self.client.patch(self._settings_url, json={"privileges": list(all_priv.values())}))
        self.get_credentials()
        priv = self.enables_by_id[enable.privilege_id]
        logger.info(f"Updated enable privilege with username {priv.username} and ID of {priv.privilege_id}")
        return priv

    def create_enable(
        self,
        username: str,
        password: str,
        networks: list = None,
        notes: str = None,
        excluded: list = None,
        expiration: str = None,
    ) -> Privilege:
        """
        Creates a new enable password (privilege account).
        Requires username and password and will default to all networks with no expiration.
        After creation Authentication.enables will be updated with new priorities and the new enable is returned.
        :param username: str: Username
        :param password: str: Unencrypted password
        :param networks: list: List of networks defaults to ["0.0.0.0/0"]
        :param notes: str: Optional Note/Description of credential
        :param excluded: list: Optional list of networks to exclude
        :param expiration: str: Optional date for expiration, if none then do not expire.
                                To ensure correct date use YYYYMMDD or MM/DD/YYYY formats
        :return: Privilege: Obj: Privilege Obj with ID and encrypted password
        """
        if self.snapshot_id:
            raise NotImplementedError(CRED_ERROR)
        payload = self._create_payload(username, password, notes, networks, excluded, expiration)
        payload["includeNetworks"] = payload.pop("network")
        res = raise_for_status(self.client.post(self._priv_url, json=payload))
        self.get_enables()
        priv = Privilege(**res.json())
        logger.info(f"Created enable password with username {priv.username} and ID of {priv.privilege_id}")
        return priv

    def delete_credential(self, credential: Union[Credential, str]) -> bool:
        """
        Deletes a credential and updates Authentication.credentials with new priorities.
        :param credential: Union[Credential, str]: Cred ID in a string or Credential object
        :return:
        """
        if self.snapshot_id:
            raise NotImplementedError(CRED_ERROR)
        cred = credential.credential_id if isinstance(credential, Credential) else credential
        raise_for_status(self.client.request("DELETE", self._cred_url, json=[cred]))
        self.get_credentials()
        logger.warning(f"Deleted credential ID {cred}")
        return True

    def delete_enable(self, enable: Union[Privilege, str]) -> bool:
        """
        Deletes an enable password (privilege account) and updates Authentication.enable with new priorities.
        :param enable: Union[Privilege, str]: Enable ID in a string or Privilege object
        :return:
        """
        if self.snapshot_id:
            raise NotImplementedError(CRED_ERROR)
        priv = enable.privilege_id if isinstance(enable, Privilege) else enable
        raise_for_status(self.client.request("DELETE", self._priv_url, json=[priv]))
        self.get_enables()
        logger.warning(f"Deleted enable password ID {priv}")
        return True

    def update_cred_priority(self, credentials: dict) -> dict:
        """
        Updates the priority of credentials.  Reorder Authentication.credentials dictionary and submit to this method.
        :param credentials: dict: {priority: Credential}
        :return: self.credentials: dict: {priority: Credential}
        """
        if self.snapshot_id:
            raise NotImplementedError(CRED_ERROR)
        payload = [{"id": c.credential_id, "priority": p} for p, c in credentials.items()]
        raise_for_status(self.client.patch(self._cred_url, json=payload))
        self.get_credentials()
        return self.credentials

    def update_enable_priority(self, enables: dict) -> dict:
        """
        Updates the priority of enable passwords.  Reorder Authentication.enables dictionary and submit to this method.
        :param enables: dict: {priority: Privilege}
        :return: self.enables: dict: {priority: Privilege}
        """
        if self.snapshot_id:
            raise NotImplementedError(CRED_ERROR)
        payload = [{"id": e.privilege_id, "priority": p} for p, e in enables.items()]
        raise_for_status(self.client.patch(self._priv_url, json=payload))
        self.get_enables()
        return self.enables



import logging
import random
import re
import sys
from datetime import datetime, timezone
from importlib import metadata
from time import sleep
from typing import Optional, Union, Any, Type, Literal
from urllib.parse import urlparse, urljoin, ParseResult

import jwt
from niquests import (
    Session,
    Request,
    Response,
    PreparedRequest,
    RetryConfiguration,
    TimeoutConfiguration,
    LifeCycleHook,
)

# See https://github.com/jawah/niquests/issues/324 for the PathLike import
from niquests.typing import TLSVerifyType, TimeoutType, ProxyType, TLSClientCertType, PathLike  # noqa: F401
from niquests.auth import AuthBase
from niquests.cookies import extract_cookies_to_jar
from niquests.exceptions import RetryError, HTTPError
from packaging.version import Version
from pydantic import field_validator, Field, AliasChoices, PrivateAttr, AnyUrl
from pydantic_settings import BaseSettings, SettingsConfigDict, PydanticBaseSettingsSource, InitSettingsSource

from ipfabric.tools.shared import raise_for_status, valid_snapshot

logger = logging.getLogger("ipfabric")

RE_VERSION = re.compile(r"v?(\d(\.\d*)?)")


def log_request(request: PreparedRequest):
    # TODO: Figure out why these don't print
    logger.debug(f"Request event hook: {request.method} {request.url} - Waiting for response")


def log_response(response: Response):
    request = response.request
    logger.debug(f"Response event hook: {request.method} {request.url} - Status {response.status_code}")


class RateLimiter(AuthBase):
    def __init__(self, api_version: str = None):
        self.os_api_version = api_version

    def __call__(self, r: PreparedRequest) -> PreparedRequest:
        """Add the rate limit handler to the request."""
        if self.rate_limit not in r.hooks["response"]:
            r.register_hook("response", self.rate_limit)
        if self.deprecated not in r.hooks["response"]:
            r.register_hook("response", self.deprecated)
        return r

    def deprecated(self, response: Response):
        if response.headers.get("deprecation", False) == "true":
            path = urlparse(response.url).path
            sunset = response.headers.get("sunset", None)
            if path.split("/")[2] == self.os_api_version:
                msg = f"API endpoint '{path}' has deprecation header set and will be removed in a future release."
                msg += f" Sunset date is {sunset}." if sunset else ""
                logger.warning(msg)
            else:
                # TODO: Change to new versioning scheme.
                logger.info(
                    f"API endpoint '{path}' is using older API version, current IP Fabric version is {self.os_api_version}."
                )

    @staticmethod
    def send_prepared_request(response: Response, prep: PreparedRequest, **kwargs) -> Response:
        _r = response.connection.send(prep, **kwargs)
        _r.history = [*response.history, response]
        _r.request = prep
        return _r

    def prepare_and_send(self, response: Response, **kwargs) -> Response:
        response.close()
        prep = response.request.copy()
        extract_cookies_to_jar(prep._cookies, response.request, response.raw)
        prep.prepare_cookies(prep._cookies)
        return self.send_prepared_request(response, prep, **kwargs)

    def rate_limit(self, response: Response, **kwargs):
        if response.status_code != 429:
            return response
        retry = response.connection.max_retries

        reset = int(response.headers.get("X-RateLimit-Reset", 0))
        _ = retry.backoff_jitter * random.random()  # nosemgrep: bandit.B311
        wait = reset - int(datetime.now(timezone.utc).timestamp()) + _

        retries = len([_ for _ in response.history if _.status_code in retry.status_forcelist])
        if retries >= retry.status:
            raise RetryError("Max retry limit reached", response=response)
        if wait > 0:
            print(f"Rate Limit Reached. Waiting for {wait} seconds.")
            sleep(wait)
            _r = self.prepare_and_send(response, **kwargs)
            if _r.status_code == 429 and _r.headers.get("X-RateLimit-Remaining", None) == "0":
                self.rate_limit(_r, **kwargs)

            return _r
        return response


class JWTToken(RateLimiter):
    """Used for streamlit."""

    def __call__(self, r: PreparedRequest) -> PreparedRequest:
        """Add the token to the request headers."""
        super().__call__(r)
        if self.auth_flow not in r.hooks["response"]:
            r.register_hook("response", self.auth_flow)
        return r

    @staticmethod
    def auth_flow(response: Response) -> Response:
        if response.status_code == 401:
            logger.warning("Access Token has expired, please refresh IP Fabric or log in again.")
        return response


class TokenAuth(RateLimiter):
    def __init__(self, token: str, api_version: str = None) -> None:
        super().__init__(api_version)
        self._auth_header = token

    def __call__(self, r: PreparedRequest) -> PreparedRequest:
        """Add the token to the request headers."""
        super(TokenAuth, self).__call__(r)
        r.headers["X-API-Token"] = self._auth_header
        return r


class AccessToken(RateLimiter):
    def __init__(
        self,
        username: str,
        password: str,
        base_url: AnyUrl,
        api_version: str = None,
    ):
        super().__init__(api_version)
        self.base_url = base_url
        self.username = username
        self.password = password

    def __call__(self, r: PreparedRequest) -> PreparedRequest:
        """Add the token to the request headers."""
        super().__call__(r)
        if self.auth_flow not in r.hooks["response"]:
            r.register_hook("response", self.auth_flow)
        return r

    def auth_flow(self, response: Response, **kwargs) -> Response:
        if response.status_code == 401 and not response.request._cookies.get("accessToken"):
            return self.login(response, **kwargs)
        elif response.status_code == 401 and "API_EXPIRED_ACCESS_TOKEN" in response.text:
            return self.refresh(response, **kwargs)
        return response

    def login(self, response: Response, **kwargs) -> Response:
        response.close()
        auth_prep = Request(
            method="POST",
            url=urljoin(self.base_url, "auth/login"),
            json={"username": self.username, "password": self.password},
            headers={"Content-Type": "application/json"},
        ).prepare()
        auth = response.connection.send(auth_prep, **kwargs)
        if auth.status_code != 200:
            raise HTTPError("Failed to Authenticate", response=auth, request=auth_prep)

        prep = self.prepare_auth(auth, response)
        return self.send_prepared_request(response, prep, **kwargs)

    def refresh(self, response: Response, **kwargs) -> Response:
        # Use refreshToken in Cookies to get new accessToken & Response updates accessToken in shared CookieJar
        response.close()
        reauth_prep = Request(
            method="POST",
            url=urljoin(self.base_url, "/api/auth/token"),
            headers={"Content-Type": "application/json"},
        ).prepare()
        reauth = response.connection.send(reauth_prep, **kwargs)
        if reauth.status_code != 200:
            raise HTTPError("Failed to Reauthenticate", response=reauth, request=reauth_prep)
        prep = self.prepare_auth(reauth, response)
        return self.send_prepared_request(response, prep, **kwargs)

    @staticmethod
    def prepare_auth(auth: Response, response: Response) -> PreparedRequest:
        auth.close()
        response.history.append(auth)
        prep = response.request.copy()
        extract_cookies_to_jar(prep._cookies, auth.request, auth.raw)
        prep.prepare_cookies(prep._cookies)
        return prep


class MyInitSettingsSource(InitSettingsSource):
    def __init__(self, settings_cls, init_kwargs: dict[str, Any]):
        timeout = init_kwargs.pop("timeout", "DEFAULT")
        init_kwargs = {k: v for k, v in init_kwargs.items() if v is not None}
        if timeout != "DEFAULT":
            init_kwargs["timeout"] = float(timeout) if isinstance(timeout, str) else timeout
        super().__init__(settings_cls, init_kwargs)


class Setup(BaseSettings):
    model_config = SettingsConfigDict(env_file=".env", env_file_encoding="utf-8", env_prefix="ipf_", extra="allow")
    base_url: AnyUrl = Field(None, validation_alias=AliasChoices("base_url", "ipf_url"))
    api_version: Optional[Union[int, float, str]] = Field(
        None, validation_alias=AliasChoices("api_version", "ipf_version")
    )
    auth: Optional[Any] = Field(None, alias="auth", exclude=True)
    token: Optional[str] = None
    username: Optional[str] = None
    password: Optional[str] = None
    snapshot_id: Union[str, None] = Field("$last", validation_alias=AliasChoices("snapshot_id", "ipf_snapshot"))
    nvd_api_key: Optional[str] = Field(None, alias="nvd_api_key")

    verify: TLSVerifyType = True
    timeout: Union[None, TimeoutType, Literal["DEFAULT"]] = TimeoutConfiguration(connect=5.0, read=5.0)
    proxy: Optional[ProxyType] = None
    cert: Optional[TLSClientCertType] = None
    http2: bool = True
    event_hooks: Optional[dict[str, list[LifeCycleHook]]] = Field(default_factory=dict)
    retry_config: Optional[RetryConfiguration] = Field(None, description="Retries configuration for niquests client.")
    debug: bool = False
    psql: bool = False

    _client: Optional[Session] = PrivateAttr(None)
    _os_version: Optional[str] = PrivateAttr(None)
    _os_api_version: Optional[str] = PrivateAttr(None)
    _auth_type: Optional[str] = PrivateAttr(None)

    @property
    def client(self) -> Session:
        return self._client

    @classmethod
    def settings_customise_sources(
        cls,
        settings_cls: Type[BaseSettings],
        init_settings: PydanticBaseSettingsSource,
        env_settings: PydanticBaseSettingsSource,
        dotenv_settings: PydanticBaseSettingsSource,
        file_secret_settings: PydanticBaseSettingsSource,
    ) -> tuple[PydanticBaseSettingsSource, ...]:
        return (
            MyInitSettingsSource(settings_cls, init_settings.init_kwargs),
            env_settings,
            dotenv_settings,
            file_secret_settings,
        )

    def model_post_init(self, __context):
        self._client = Session(
            base_url=str(self.base_url),
            disable_http2=not self.http2,
            timeout=self.timeout,
            retries=self.retry_config,
        )

        self.client.verify = self.verify
        self.client.cert = self.cert
        self.client.proxies = self.proxy
        self.client.headers["User-Agent"] = (
            f'python-ipfabric-sdk/{metadata.version("ipfabric")} (Python {sys.version.split(" ")[0]})'
        )
        for k, v in self.event_hooks.items():
            self.client.hooks[k].extend(v)
        if self.debug:
            self.client.hooks["pre_request"].append(log_request)
            self.client.hooks["response"].append(log_response)

        if not (self.auth or self.token) and not (self.username and self.password):
            raise RuntimeError("IP Fabric Authentication not provided.")
        self.api_version, self._os_version, self._os_api_version = self.check_version(self.api_version)
        if Version(self._os_api_version[1:]) < Version("7.5"):
            self.base_url = urljoin(self.base_url, f"api/{self.api_version}/")
        else:
            self.psql = True
            self.base_url = urljoin(self.base_url, "api/")
        self.client.base_url = self.base_url
        self._verify_auth()

    def _verify_auth(self):
        if self.auth:
            self.client.auth = self._get_auth_from_auth()
        elif self.token:
            self.client.auth = self._check_jwt(self.token)
        elif self.username and self.password:
            self.client.auth = self._get_access_token_auth(self.username, self.password)

    def _check_jwt(self, token):
        try:
            jwt.decode(token, options={"verify_signature": False})  # NOSONAR
            self.client.cookies.set("accessToken", token)
            self._auth_type = "JWT_AUTH"
            return JWTToken(api_version=self._os_api_version)
        except jwt.exceptions.DecodeError:
            self._auth_type = "API_TOKEN_AUTH"
            return TokenAuth(token, api_version=self._os_api_version)

    def _get_auth_from_auth(self) -> Any:
        """Separate auth handling logic"""
        if isinstance(self.auth, str):
            return self._check_jwt(self.auth)
        elif isinstance(self.auth, tuple):
            return self._get_access_token_auth(self.auth[0], self.auth[1])
        self._auth_type = type(self.client.auth)
        return self.auth

    def _get_access_token_auth(self, username: str, password: str) -> AccessToken:
        """Separate auth handling logic"""
        self._auth_type = "USER_PASS_ACCESS_TOKEN"
        return AccessToken(
            username,
            password,
            self.base_url,
            api_version=self._os_api_version,
        )

    @property
    def update_attrs(self) -> dict:
        return {
            "base_url": self.base_url,
            "api_version": self.api_version,
            "auth": self._auth_type,
            "timeout": self.client.timeout,
            "_os_version": self._os_version,
            "_os_api_version": self._os_api_version,
            "debug": self.debug,
            "_client": self.client,
            "verify": self.verify,
            "proxy": self.proxy,
            "http2": self.http2,
            "nvd_api_key": self.nvd_api_key,
            "_psql": self.psql,
        }

    def check_version(self, custom_api_version) -> tuple:
        """Checks API Version and returns the version to use in the URL and the OS Version

        Returns:
            api_version, os_version
        """
        cfg_version = Version(custom_api_version)
        api_version = f"v{cfg_version.major}.{cfg_version.minor}"

        resp = raise_for_status(self.client.get("/api/version")).json()
        os_api_version = Version(resp["apiVersion"])

        if os_api_version.major != cfg_version.major:
            raise RuntimeError(
                f"OS Major Version `{os_api_version.major}` does not match SDK Major Version `{cfg_version.major}`."
            )

        if cfg_version.minor > os_api_version.minor:
            logger.warning(
                f"Specified SDK Version `{api_version}` is greater than "
                f"OS API Version `{os_api_version}`, using OS Version."
            )
            api_version = f"v{os_api_version.base_version}"

        return api_version, resp["releaseVersion"], resp["apiVersion"]

    @field_validator("api_version")
    @classmethod
    def _valid_version(cls, v: Union[None, int, float, str]) -> Union[None, str]:
        if not v:
            return metadata.version("ipfabric")
        re_version = RE_VERSION.match(str(v))
        if not re_version:
            raise ValueError(f"IPF_VERSION ({v}) is not valid, must be like `v#` or `v#.#`.")
        elif re_version and re_version.group(2):
            return "v" + RE_VERSION.match(str(v)).group(1)
        else:
            return "v" + RE_VERSION.match(str(v)).group(1) + ".0"

    @field_validator("base_url")
    @classmethod
    def _convert_url(cls, v: Union[ParseResult, str]) -> str:
        if isinstance(v, str):
            v = urlparse(v)
        return str(v)

    @field_validator("snapshot_id")
    @classmethod
    def _valid_snapshot(cls, v: Union[None, str]) -> str:
        return valid_snapshot(v, init=True)

    @field_validator("verify")
    @classmethod
    def _verify(cls, v: Union[bool, int, str]) -> Union[bool, str]:
        if isinstance(v, bool):
            return v
        elif isinstance(v, int):
            return bool(v)
        elif v.lower() in {"0", "off", "f", "false", "n", "no", "1", "on", "t", "true", "y", "yes"}:
            return False if v.lower() in {0, "0", "off", "f", "false", "n", "no"} else True
        else:
            return v

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Needed for context"""
        pass



import logging
from typing import Any, Optional, Union

from niquests import HTTPError
from pydantic import BaseModel, ConfigDict, Field

from .api_tokens import APIToken
from .appliance import ApplianceConfiguration
from .attributes import Attributes
from .authentication import Authentication
from .discovery import Discovery
from .local_users import LocalUsers
from .rbac import Roles, Policies
from .seeds import Seeds
from .site_separation import SiteSeparation
from .vendor_api import VendorAPI
from ipfabric.tools.shared import raise_for_status

logger = logging.getLogger("ipfabric")


class Settings(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    client: Any = Field(exclude=True)
    _local_users: Optional[Union[LocalUsers, bool]] = None
    _api_tokens: Optional[Union[APIToken, bool]] = None
    _site_separation: Optional[Union[SiteSeparation, bool]] = None
    _discovery: Optional[Union[Discovery, bool]] = None
    _authentication: Optional[Union[Authentication, bool]] = None
    _seeds: Optional[Union[Seeds, bool]] = None
    _vendor_api: Optional[Union[VendorAPI, bool]] = None
    _global_attributes: Optional[Union[Attributes, bool]] = None
    _roles: Optional[Union[Roles, bool]] = None
    _policies: Optional[Union[Policies, bool]] = None
    _appliance_configuration: Optional[Union[ApplianceConfiguration, bool]] = None

    def _try_object(self, obj, name, attr):
        if getattr(self, attr) is None:
            try:
                setattr(self, attr, obj(client=self.client))
            except HTTPError:
                setattr(self, attr, False)
        if getattr(self, attr) is not False:
            return getattr(self, attr)
        logger.warning(f"{self.client._api_insuf_rights} for `{name}` class.")
        return None

    @property
    def local_users(self) -> LocalUsers:
        return self._try_object(LocalUsers, "LocalUsers", "_local_users")

    @property
    def roles(self) -> Roles:
        return self._try_object(Roles, "Roles", "_roles")

    @property
    def policies(self) -> Policies:
        return self._try_object(Policies, "Policies", "_policies")

    @property
    def api_tokens(self) -> APIToken:
        return self._try_object(APIToken, "APIToken", "_api_tokens")

    @property
    def site_separation(self) -> SiteSeparation:
        return self._try_object(SiteSeparation, "SiteSeparation", "_site_separation")

    @property
    def discovery(self) -> Union[Discovery, None]:
        if self._discovery is None:
            try:
                settings = raise_for_status(self.client.get("settings")).json()
                self._discovery = Discovery(
                    client=self.client,
                    vendorApi=VendorAPI(client=self.client, vendor_api=settings.pop("vendorApi")),
                    authentication=Authentication(
                        client=self.client,
                        settings={"credentials": settings.pop("credentials"), "privileges": settings.pop("privileges")}
                    ),
                    **settings,
                )
            except HTTPError:
                self._discovery = False
        if self._discovery is not False:
            return self._discovery
        logger.warning(f"{self.client._api_insuf_rights} for `Discovery` class.")
        return None

    @property
    def authentication(self) -> Authentication:
        return self._try_object(Authentication, "Authentication", "_authentication")

    @property
    def seeds(self) -> Seeds:
        return self._try_object(Seeds, "Seeds", "_seeds")

    @property
    def vendor_api(self) -> VendorAPI:
        return self._try_object(VendorAPI, "VendorAPI", "_vendor_api")

    @property
    def global_attributes(self) -> Attributes:
        if self._global_attributes is None:
            self._global_attributes = Attributes(client=self.client)
        return self._global_attributes

    @property
    def appliance_configuration(self) -> ApplianceConfiguration:
        return self._try_object(ApplianceConfiguration, "ApplianceConfiguration", "_appliance_configuration")



from .api_tokens import APIToken
from .attributes import Attributes
from .authentication import Authentication
from .discovery import Discovery
from .local_users import LocalUsers
from .seeds import Seeds
from ipfabric.models.discovery import SeedList, Networks  # TODO: 8.0 remove
from .settings import Settings
from .site_separation import SiteSeparation
from .vendor_api import VendorAPI

__all__ = [
    "APIToken",
    "Attributes",
    "Authentication",
    "Seeds",
    "SeedList",
    "SiteSeparation",
    "LocalUsers",
    "VendorAPI",
    "Discovery",
    "Networks",
    "Settings",
]



from functools import cached_property
from typing import Union

from pydantic import BaseModel, Field


class BaseFilter(BaseModel):
    value: Union[str, bool, int, float] = Field(description="The value to use for the filter.", title="Filter Value")
    operator: str = Field(
        "eq",
        description="The IP Fabric operator to use for the filter (see API documentation).",
        title="Filter Operator",
        examples=[
            "eq",
            "neq",
            "ieq",
            "nieq",
            "lt",
            "lte",
            "gt",
            "gte",
            "like",
            "notlike",
            "reg",
            "nreg",
            "ireg",
            "nireg",
        ],
    )


class AttributeFilter(BaseFilter, BaseModel):
    key: str = Field(
        description="The Attribute Name to filter on.", title="Attribute Name", examples=["MGMT_IP", "COUNTRY"]
    )

    @cached_property
    def format(self) -> dict[str, list]:
        """Format the filter for use in the API request."""
        return {"device.attributes": [self.key, self.operator, self.value]}


class Filter(BaseFilter, BaseModel):
    column: str = Field(
        description="The name of the column to use for the filter.",
        title="API Column Name",
        examples=["primaryIp", "hostname"],
    )

    @cached_property
    def format(self) -> dict[str, list]:
        """Format the filter for use in the API request."""
        return {self.column: [self.operator, self.value]}


class DeviceFilter(Filter, BaseModel):
    @cached_property
    def format(self) -> dict[str, list]:
        """Format the filter for use in the API request."""
        return {f"device.{self.column}": [self.operator, self.value]}



import hmac
import sys
from importlib.util import spec_from_file_location, module_from_spec
from pathlib import Path
from typing import Optional, Union, Annotated, Any

from fastapi import Request, status, BackgroundTasks, Response
from pydantic import Field, BaseModel, field_validator
from pydantic_settings import BaseSettings, SettingsConfigDict
from sqlalchemy import func
from sqlmodel import Session, select, or_

from ipfabric.models.webhooks import IntentEvent, SnapshotEvent, WebhookEvent

AnnotatedWebhookEvent = Annotated[Union[IntentEvent, SnapshotEvent], Field(discriminator="type")]


class WebhookResponse(BaseModel):
    total: int
    size: int
    start: int
    limit: int
    next: Optional[str] = None
    items: list[WebhookEvent]


class WebhookConfig(BaseSettings, cli_parse_args=True):
    model_config = SettingsConfigDict(env_file=".env", env_file_encoding="utf-8", extra="allow")

    # uvicorn settings
    uvicorn_ip: str = Field("0.0.0.0", description="IP address to bind the FastAPI server.")
    uvicorn_port: int = Field(8000, description="Port to bind the FastAPI server.")
    uvicorn_reload: bool = Field(False, description="Enable auto-reload for the FastAPI server.")

    # API settings
    ipf_secret: Optional[str] = Field(
        None, description="Secret token to validate incoming webhook requests. Validation will be skipped if not set."
    )
    database_url: str = Field("sqlite:///database.db", description="Database URL for storing webhook events.")
    title: str = Field("IP Fabric Webhook Receiver", description="Title of the FastAPI application.")
    webhook_path: str = Field(
        "/ipfabric", description="Path where the webhook endpoint will be available, default '/ipfabric'."
    )
    custom_settings: Optional[Any] = Field(
        None, description="Settings to pass to your custom functions. You functions must handle this parameter."
    )

    # Snapshot event settings
    snapshot_method: Optional[str] = Field(
        None,
        description="Path to the Python file containing the snapshot processing method in 'path:function' format.",
        examples=["my_script.py:snapshot_event"],
    )
    snapshot_scheduled_only: bool = Field(
        True, description="If set to True, only process webhooks from scheduled snapshots"
    )

    # Intent verification event settings
    intent_method: Optional[str] = Field(
        None,
        description="Path to the Python file containing the intent processing method in 'path:function' format.",
        examples=["my_script.py:intent_event"],
    )
    intent_scheduled_only: bool = Field(
        True,
        description="If set to True, only process webhooks from scheduled snapshots' intent verifications. "
        "Requires 'snapshot_event' to be configured and received for intent calculations.",
    )

    @field_validator("snapshot_method", "intent_method", mode="after")
    @classmethod
    def load_methods(cls, value: Optional[str]):
        if not value:
            return value
        _ = value.split(":")
        if len(_) != 2:
            raise ValueError("Method must be in 'path:function' format.")
        path = Path(_[0])
        if not path.is_file():
            raise ValueError(f"File '{str(path)}' does not exist.")
        module_name = str(path)
        if module_name not in sys.modules:
            spec = spec_from_file_location(module_name, path)
            module = module_from_spec(spec)
            sys.modules[module_name] = module
            spec.loader.exec_module(module)
        else:
            module = sys.modules[module_name]
        if not hasattr(module, _[1]):
            raise ValueError(f"Function '{_[1]}' not found in module '{module_name}'.")
        return getattr(module, _[1])


async def verify_signature(request: Request, x_ipf_signature: str, secret: Optional[str] = None) -> tuple[bool, str]:
    """Verify the HMAC signature of the incoming request"""
    if secret:
        if x_ipf_signature is None:
            return False, "X-IPF-Signature header missing."

        input_hmac = hmac.new(key=secret.encode(), msg=await request.body(), digestmod="sha256")
        if not hmac.compare_digest(input_hmac.hexdigest(), x_ipf_signature):
            # Signature does not match, possible tampering
            return False, "X-IPF-Signature does not match."
    return True, ""


def calc_method(event: AnnotatedWebhookEvent, bg_tasks: BackgroundTasks, session: Session, settings: WebhookConfig):
    """Determine and execute the appropriate method for the event"""
    etype = "snapshot" if event.type == "snapshot" else "intent"
    if not getattr(settings, f"{etype}_event"):
        # If event type processing is disabled, return 204 No Content
        return Response(status_code=status.HTTP_204_NO_CONTENT)

    # Store the event in the database
    record = WebhookEvent.model_validate(event)
    session.add(record)
    session.commit()
    session.refresh(record)

    if etype == "snapshot" and settings.snapshot_scheduled_only and event.requester != "cron":
        # If not a scheduled snapshot, skip processing
        return record

    # Check if the related snapshot was a scheduled one
    query = (
        select(WebhookEvent)
        .where(WebhookEvent.type == "snapshot")
        .where(WebhookEvent.action == "discover")
        .where(WebhookEvent.snapshotId == event.snapshotId)
        .where(WebhookEvent.status == "completed")
    )
    if (
        etype == "intent"
        and settings.intent_scheduled_only
        and (
            event.requester != "snapshot:discover"  # Intent not from snapshot discover
            or not session.exec(query.where(WebhookEvent.requester == "cron")).first()  # No scheduled snapshot found
            or session.exec(query.where(WebhookEvent.requester != "cron")).first()  # Snapshot was modified by user
        )
    ):
        # If not a scheduled snapshot's (or modified scheduled snapshot) intent verification, skip processing
        return record

    if method := getattr(settings, f"{etype}_method"):
        bg_tasks.add_task(method, event=record, settings=settings.custom_settings)
    return record


def query_db(
    session: Session,
    endpoint: Optional[str] = None,
    offset: int = 0,
    limit: int = 100,
    snapshotId: Optional[str] = None,
):
    """Generic database query function"""
    query = or_(WebhookEvent.type == "snapshot", WebhookEvent.type == "intent-verification")
    if endpoint == "snapshots":
        query = WebhookEvent.type == "snapshot"
    elif endpoint == "intents":
        query = WebhookEvent.type == "intent-verification"
    if snapshotId:
        count = session.exec(
            select(func.count(WebhookEvent.dbId)).where(query).where(WebhookEvent.snapshotId == snapshotId)
        ).one()
        items = session.exec(
            select(WebhookEvent).where(query).where(WebhookEvent.snapshotId == snapshotId).offset(offset).limit(limit)
        ).all()
    else:
        count = session.exec(select(func.count(WebhookEvent.dbId)).where(query)).one()
        items = session.exec(select(WebhookEvent).where(query).offset(offset).limit(limit)).all()
    link = f"/{endpoint}/?offset={offset + limit}&limit={limit}" if count > len(items) + offset else None
    if link and snapshotId:
        link += f"&snapshotId={snapshotId}"
    return {
        "total": count,
        "size": len(items),
        "items": items,
        "start": offset,
        "limit": limit,
        "next": link,
    }



from functools import cached_property
from typing import Optional, Union, ForwardRef, Any

from pydantic import BaseModel, Field, model_validator, PrivateAttr, computed_field, ConfigDict

from ipfabric.dynamic_attributes.configs.filters import Filter, AttributeFilter, DeviceFilter
from ipfabric.dynamic_attributes.configs.values import TableValue, ConfigValue

REPORT_COLS = [
    "attribute",
    "global_value",
    "local_value",
    "no_config",
    "name",
    "api_endpoint",
    "no_start_config",
    "no_current_config",
    "new_value",
    "column_value",
    "static",
    "null_value",
    "regex_pattern",
    "regex_group",
    "regex_match",
    "mapping_match",
    "delete",
    "overwrite",
]

TableRule, ConfigRule = ForwardRef("TableRule"), ForwardRef("ConfigRule")
DefaultRule, DefaultConfigRule = ForwardRef("DefaultRule"), ForwardRef("DefaultConfigRule")


def merge_default(
    original: Union[TableRule, ConfigRule], default: Union[DefaultRule, DefaultConfigRule]
) -> Union[TableRule, ConfigRule]:
    if not default:
        original.overwrite = original.overwrite or False
        return original
    if default.overwrite is not None and original.overwrite is None:
        # Copy default overwrite to Rule:
        original.overwrite = default.overwrite
    # If Rule does not have overwrite set, set it to False:
    original.overwrite = original.overwrite or False
    # Copy Default Attribute Name to Rule:
    original.attribute = original.attribute or default.attribute
    if default.delete_attribute is not None and original.delete_attribute is None:
        # Copy Default Delete Attribute to Rule if Rule does not have delete_attribute set:
        original.delete_attribute = default.delete_attribute
    if original.merge_default_filters:
        if default.filter_string and not original.filters:
            # Copy Default Filter String to Rule if Rule does not have filters and merge_default_filters is True:
            original.filter_string = default.filter_string
        if default.filters and not original.filter_string:
            # Copy Default Filters to Rule if Rule does not have a filter_string and merge_default_filters is True:
            original.filters.extend(default.filters)
    return original


class FilterRule(BaseModel):
    filters: list[Filter] = Field(
        default_factory=list, description="List of filters to apply to table endpoint.", title="Table Filters"
    )
    filter_string: Optional[str] = Field(
        None,
        description="A string to use for the filter on table endpoint instead of the filters list.",
        title="Table Filter String",
        examples=['{"vendor":["like","cisco"]}'],
    )

    @model_validator(mode="after")
    def _validate(self):
        if self.filters and self.filter_string:
            raise ValueError("'rule[filter_string]' cannot be combined with 'rule[filters]'.")
        return self

    @property
    def filter_list(self) -> list[dict]:
        return [_.format for _ in self.filters]

    @property
    def calc_filter(self) -> Union[dict[str, list[dict]], str]:
        if self.filter_string:
            return self.filter_string
        return {"and": self.filter_list} if self.filter_list else {}

    @property
    def filter_hash(self) -> Union[int, None]:
        """Return a hash of the filter string or list."""
        if self.filter_string:
            return hash(self.filter_string)
        return hash(str(sorted(self.filter_list, key=lambda x: str(x)))) if self.filter_list else None


class BaseRule(FilterRule, BaseModel):
    attribute_filters: list[AttributeFilter] = Field(
        default_factory=list, description="Attribute Filters to apply to table endpoint.", title="Attribute Filters"
    )

    @model_validator(mode="after")
    def _validate(self):
        if (self.filters or self.attribute_filters) and self.filter_string:
            raise ValueError(
                "'rule[filter_string]' cannot be combined with 'rule[filters]' or 'rule[attribute_filters]'."
            )
        return self

    @property
    def filter_list(self) -> list[dict]:
        return [*super().filter_list, *[_.format for _ in self.attribute_filters]]


class InventoryRule(BaseRule, BaseModel):
    report_columns: list[str] = Field(
        default_factory=list,
        description="List of Inventory > Devices columns to also include in the Sync Report; "
        "the report will always return [hostname, sn, siteName].",
        title="Inventory Report Columns",
        examples=["model"],
    )

    @cached_property
    def calc_columns(self) -> list[str]:
        """Return the columns to use for the Inventory."""
        return [_ for _ in {"hostname", "sn", "siteName", *self.report_columns}]

    @cached_property
    def df_columns(self) -> list[str]:
        """Return the columns to use for the DataFrame."""
        return [*self.calc_columns, *REPORT_COLS]


class CommonDefaultRule(BaseModel):
    """Common Default Rule for Table and Config Rules."""

    attribute: Optional[str] = Field(
        None,
        description="The name of the Attribute to be updated or created.",
        title="Attribute Name",
        examples=["COUNTRY", "MGMT_IP", "ENVIRONMENT"],
    )
    delete_attribute: Optional[bool] = Field(
        None,
        description="Caution: If no match is found this will delete the Attribute from all devices in the filtered "
        "inventory list for this Attribute Name. After running each rule and no match is found, the "
        "attribute will be flagged for deletion in the remaining inventory list. If a later rule matches "
        "does find a match then the attribute will not be deleted and the new value will be assigned.",
        title="Delete Attribute (DANGEROUS - Read the Description)",
    )
    overwrite: Optional[bool] = Field(
        None,
        description="Overwrite existing Global Attribute values; default False.",
        title="Overwrite Global Attribute",
    )


class DefaultRule(CommonDefaultRule, BaseRule, BaseModel):
    value: Optional[TableValue] = Field(
        None,
        description="The value to assign to the Attribute, "
        "either configured in the Default Table rules or per Table rule.",
        title="Attribute Value",
    )
    device_filters: list[DeviceFilter] = Field(
        default_factory=list,
        description="Device filter to apply to the table endpoint (i.e. vendor, family, etc.).",
        title="Device Filters",
    )

    @model_validator(mode="after")
    def _validate(self):
        if (self.filters or self.attribute_filters or self.device_filters) and self.filter_string:
            raise ValueError(
                "'rule[filter_string]' cannot be combined with 'rule[filters]', "
                "'rule[attribute_filters]', or 'rule[device_filters]'."
            )
        return self

    @property
    def filter_list(self) -> list[dict]:
        return [*super().filter_list, *[_.format for _ in self.device_filters]]


class CommonRule(BaseModel):
    model_config = ConfigDict(extra="allow")
    name: str = Field(
        description="A unique name of the rule for reporting purposes.",
        title="Rule Name",
        examples=["Rule 1", "Loopback0 Rule"],
    )
    _inventory: Optional[InventoryRule] = PrivateAttr(None)

    def model_post_init(self, context: Any, /) -> None:
        self._inventory = InventoryRule(**self.model_extra["inventory"]) if "inventory" in self.model_extra else None

    @computed_field(
        title="Internal Inventory Rule",
        description="Internal Inventory Rule used for filtering devices when joining multiple configuration files.",
    )
    @property
    def inventory(self) -> Union[InventoryRule, None]:
        return self._inventory

    @inventory.setter
    def inventory(self, value: Union[InventoryRule, None]):
        self._inventory = value


class TableRule(CommonRule, DefaultRule, BaseModel):
    merge_default_filters: bool = Field(
        True,
        description="Merge 'default[filters]' with Table TableRule filters or "
        "'default[filter_string]' to Table TableRule filter_string.",
        title="Merge Default Filters",
    )
    merge_default_device_filters: bool = Field(
        True,
        description="Merge 'default[device_filters]' with Table TableRule device_filters.",
        title="Merge Default Device Filters",
    )
    merge_default_attribute_filters: bool = Field(
        True,
        description="Merge 'default[attribute_filters]' with Table TableRule attribute_filters.",
        title="Merge Default Attribute Filters",
    )

    def merge_default(self, default: DefaultRule) -> "TableRule":
        if not isinstance(default, DefaultRule):
            return self
        rule = merge_default(self.model_copy(deep=True), default)
        # Copy Default Value Config to TableRule:
        rule.value = rule.value or default.value
        # Copy Default SN Column to TableRule:
        rule.value.sn_column = rule.value.sn_column or (default.value.sn_column if default.value else "sn")
        if not rule.filter_string:
            if default.device_filters and rule.merge_default_device_filters:
                # Copy Default Device Filters to TableRule if merge_default_filters is True and filter_string is not set:
                rule.device_filters.extend(default.device_filters)
            if default.attribute_filters and rule.merge_default_attribute_filters:
                # Copy Default Attribute Filters to TableRule if merge_default_attribute_filters is True:
                rule.attribute_filters.extend(default.attribute_filters)
        return rule


class DefaultConfigRule(CommonDefaultRule, FilterRule, BaseModel):
    filters: list[Filter] = Field(
        default_factory=list,
        description="Additional Inventory Filters to apply to limit devices for configuration matching.",
        title="Additional Inventory Filters",
    )
    filter_string: Optional[str] = Field(
        None,
        description="A string to use for additional Inventory Filtering instead of the filters list.",
        title="Additional Inventory Filter String",
        examples=['{"vendor":["like","cisco"]}'],
    )
    overwrite_unsupported: Optional[bool] = Field(
        None,
        description="If a device does not support configuration tasks and a custom Attribute is set then the default"
        " (false) is to ignore overwriting the manual Attribute with the 'no_config_value' in the value "
        "config. This provides more flexibility than the 'overwrite' setting which is only used for "
        "supported devices. If this is set to 'true' then the 'no_config_value' will be used to "
        "overwrite the manual Attribute value.",
        title="Overwrite Unsupported Devices' Attributes",
    )

    @model_validator(mode="after")
    def _validate(self):
        if self.filters and self.filter_string:
            raise ValueError("'rule[filter_string]' cannot be combined with 'rule[filters]'.")
        return self


class ConfigRule(CommonRule, DefaultConfigRule, BaseModel):
    value: ConfigValue = Field(
        description="The value to assign to the Attribute based on a configuration search.",
        title="Attribute Value",
    )
    merge_default_filters: bool = Field(
        True,
        description="Merge 'default[config_filters]' with Configuration TableRule filters.",
        title="Merge Default Configuration Filters",
    )

    def merge_default(self, default: DefaultConfigRule) -> "ConfigRule":
        """Merge Default Config TableRule with Config TableRule"""
        if isinstance(default, DefaultConfigRule):
            rule = merge_default(self.model_copy(deep=True), default)
            if default.overwrite_unsupported is not None and rule.overwrite_unsupported is None:
                # Copy default overwrite_unsupported to Rule:
                rule.overwrite_unsupported = default.overwrite_unsupported
        else:
            rule = self.model_copy(deep=True)
        rule.overwrite_unsupported = rule.overwrite_unsupported or False
        return rule


TableRule.model_rebuild()
ConfigRule.model_rebuild()
DefaultRule.model_rebuild()
DefaultConfigRule.model_rebuild()



import argparse
import sys
from pathlib import Path
from typing import Union, Optional

try:
    import yaml
    import pandas  # noqa: F401
    import yaml_include  # noqa: F401
except ImportError:
    raise ImportError(
        "Required packages are missing to run this script. "
        "Please install them with 'pip3 install ipfabric[dynamic-attributes]'."
    )

from ipfabric.dynamic_attributes import AttributeSync, MultiConfig
from ipfabric.dynamic_attributes.configs import IPFabric, load_config


def _check_excel(outfile: Path) -> str:
    if not outfile:
        raise SyntaxError("Output file '-o|--outfile' must be specified with Excel format.")
    if outfile.suffix != ".xlsx":
        raise SyntaxError("Output file for Excel format must have '.xlsx' extension.")
    try:
        import openpyxl  # noqa: F401

        return "openpyxl"
    except ImportError:
        pass
    try:
        import xlsxwriter  # noqa: F401

        return "xlsxwriter"
    except ImportError:
        raise ImportError(
            "Excel format requires either 'xlsxwriter' or 'openpyxl' to be installed. "
            "Please install one of them using pip, recommended to use 'xlsxwriter'."
        )


def parse_overrides(value: Optional[Union[str, int]]) -> Union[bool, None]:
    if isinstance(value, str):
        if value.lower() in ["true", "1"]:
            return True
        elif value.lower() in ["false", "0"]:
            return False
    elif isinstance(value, int):
        return bool(value)
    return value


def parse_ipfabric_config(files: set[Path], file: Optional[Path] = None) -> Union[IPFabric, None]:
    if not file:
        return None
    if file in files or file.exists():
        return load_config(file).ipfabric
    filenames = {f.name: f for f in files}
    if len(filenames) != len(files) or file.name not in filenames:
        raise ValueError(
            f"Could not determine location of config file '{file}'; please check name or provide full path."
        )
    return load_config(filenames[file.name]).ipfabric


def main():
    arg_parser = argparse.ArgumentParser(
        description="IP Fabric Dynamic Attribute.",
        formatter_class=argparse.RawTextHelpFormatter,
        epilog="""
    This script will run the AttributeSync with the provided configuration file(s) which defaults to 'config.yml'.
    You can specify a different or multiple configuration files by passing the filename as an argument:
    ipf_dynamic_attributes mgmt-ip.yml region.yml
    """,
    )
    arg_parser.add_argument(
        "filenames",
        nargs="*",
        default=["config.yml"],
        type=Path,
        help="Configuration filename(s), defaults to 'config.yml'.",
    )
    arg_parser.add_argument(
        "-f",
        "--format",
        choices=["csv", "json", "excel"],
        default="csv",
        help="Output format for the report. Default is 'csv'. Use 'json' for JSON output.",
    )
    arg_parser.add_argument(
        "-o",
        "--outfile",
        type=Path,
        help="Output filename to send report instead of standard out.",
    )
    arg_parser.add_argument(
        "-m",
        "--merge-only",
        action="store_true",
        default=False,
        help="Merge the default rule settings into rules and display the resulting file; does not run any automation.\n"
        "This will also merge multiple configuration files into a single file.",
    )
    arg_parser.add_argument(
        "-d",
        "--dry-run-override",
        choices=["true", "false", "0", "1"],
        default=None,
        help="Override 'dry_run' setting in configuration(s); defaults to null to use value(s) in the config.\n"
        "If specifying multiple config files, all 'dry_run' values must be the same unless this flag is set.",
    )
    arg_parser.add_argument(
        "-u",
        "--update-snapshot-override",
        choices=["true", "false", "0", "1"],
        default=None,
        help="Override 'update_snapshot' setting in configuration(s); defaults to null to use value(s) in the config.\n"
        "If specifying multiple config files, all 'update_snapshot' values must be the same unless this flag is set.",
    )
    arg_parser.add_argument(
        "-i",
        "--ipfabric-override",
        type=Path,
        default=None,
        help="Override 'ipfabric' setting in configurations; defaults to null to use value in the config file(s).\n"
        "If specifying multiple config files, all 'ipfabric' values must be the same unless this flag is set "
        "to the name of the file to use the IP Fabric configuration settings from.",
    )
    args = arg_parser.parse_args()
    for file in args.filenames:
        if not file.exists():
            raise FileNotFoundError(f"Configuration file '{file}' does not exist.")

    ipf = parse_ipfabric_config(set(args.filenames), args.ipfabric_override)
    multiconfig = MultiConfig(
        configs=args.filenames,
        dry_run_override=parse_overrides(args.dry_run_override),
        update_snapshot_override=parse_overrides(args.update_snapshot_override),
        ipfabric_override=ipf,
    )

    sync = AttributeSync(config=multiconfig)

    if args.merge_only:
        print(yaml.dump(sync.config.model_dump_merged()))
        exit(0)

    engine = None
    if args.format == "excel":
        engine = _check_excel(args.outfile)

    report = sync.run()

    outfile = args.outfile or sys.stdout
    columns = [*sync.config.inventory.df_columns, "correct", "update", "create"]
    if args.format == "json":
        report.to_json(outfile, index=False, orient="records")
    elif args.format == "csv":
        report.to_csv(outfile, index=False, columns=columns)
    else:
        report.to_excel(outfile, index=False, columns=columns, engine=engine)


if __name__ == "__main__":
    main()



import logging
from typing import Any, Literal, Optional

from pydantic import BaseModel, Field, field_validator, model_serializer

from ipfabric.models.discovery import (
    Community,
    Networks,
    CliRetryLimit,
    CliSessionsLimit,
    SeedList,
    DiscoveryHistorySeeds,
    DiscoveryTask,
    BGPLimit,
    LimitDiscoveryTasks,
    ManualLink,
    ResolveNames,
    Scanner,
    Timeouts,
    Traceroute,
    Ports,
)
from ipfabric.models.site_separation import SiteSeparation
from ipfabric.settings.authentication import Authentication
from ipfabric.settings.vendor_api import VendorAPI
from ipfabric.tools.shared import raise_for_status

logger = logging.getLogger("ipfabric")


class Discovery(BaseModel):
    client: Any = Field(exclude=True)
    snapshot_id: Optional[str] = Field(None, exclude=True)
    allowTelnet: bool
    authentication: Authentication
    bgps: list[Community] = Field(default_factory=list)
    cliRetryLimit: CliRetryLimit
    cliSessionsLimit: CliSessionsLimit
    disabledPostDiscoveryActions: list[Literal["graphCache", "historicalData", "intentVerification"]] = Field(
        default_factory=list,
    )
    discoveryHistorySeeds: DiscoveryHistorySeeds
    discoveryTasks: list[DiscoveryTask] = Field(default_factory=list)
    fullBgpIpv6Limit: BGPLimit
    fullBgpLimit: BGPLimit
    limitDiscoveryTasks: LimitDiscoveryTasks
    manualLinks: list[ManualLink] = Field(default_factory=list)
    networks: Networks
    ports: list[Ports] = Field(default_factory=list)
    resolveNames: ResolveNames
    scanner: Scanner
    seedList: list[str] = Field(default_factory=list)
    siteSeparation: SiteSeparation
    timeouts: Timeouts
    traceroute: Traceroute
    vendorApi: VendorAPI

    @property
    def _endpoint(self) -> str:
        return f"snapshots/{self.snapshot_id}/settings" if self.snapshot_id else "settings"

    @field_validator("seedList")
    @classmethod
    def _verify_seed_list(cls, v: list[str]) -> list[str]:
        return SeedList(seeds=v).model_dump()

    @model_serializer
    def _ser_model(self) -> dict[str, Any]:
        ser = {}
        for k, v in self.model_fields.items():
            if v.exclude:
                continue
            if k == "authentication":
                ser.update(self.authentication.model_dump(by_alias=True))
            elif k in ["allowTelnet", "disabledPostDiscoveryActions", "seedList"]:
                ser.update({k: getattr(self, k)})
            elif k in ["bgps", "discoveryTasks", "ports", "manualLinks"]:
                ser.update({k: [_.model_dump(by_alias=True) for _ in getattr(self, k)]})
            else:
                ser.update({k: getattr(self, k).model_dump(by_alias=True)})
        return ser

    def update_discovery_networks(self, subnets: list, include: bool = False):
        payload = {}
        payload["networks"] = {}
        if include:
            payload["networks"]["include"] = subnets
            payload["networks"]["exclude"] = self.networks.exclude
        else:
            payload["networks"]["exclude"] = subnets
            payload["networks"]["include"] = self.networks.include
        res = raise_for_status(self.client.patch(self._endpoint, json=payload))
        return Networks(**res.json()["networks"])



import json
import logging
from typing import Union, Optional, Literal, Any
from urllib.parse import urljoin

from pydantic import BaseModel

from ipfabric.models import Device
from ipfabric.models.security import DefaultAction
from ipfabric.tools.shared import raise_for_status, api_header
from .input_models import (
    Unicast,
    Multicast,
    Host2GW,
    Network,
    NetworkSettings,
    PathLookupSettings,
    Overlay,
    SharedView,
    GroupSettings,
)
from .output_models.graph_result import (
    GraphResult,
    Position,
    PathLookupResult,
    TopologyResult,
)

logger = logging.getLogger("ipfabric")
GRAPHS_URL = "graphs/"
GRAPH_TYPES = Union[Unicast, Multicast, Host2GW, Network]
GRAPH_SETTINGS = Optional[Union[NetworkSettings, PathLookupSettings]]
OVERLAY_SETTINGS = Optional[Union[Overlay, dict]]
ATTRIBUTE_FILTERS = Optional[dict[str, list[str]]]
POSITION_SETTINGS = Optional[dict[str, Union[Position, dict]]]
GRAPH_RESULTS = Union[PathLookupResult, TopologyResult]


class GraphDevice(BaseModel):
    device: Device
    decidingPolicyName: Optional[str] = None
    decidingRule: Optional[list] = None
    securityType: Optional[str] = None
    securityModelRule: Optional[Union[list, DefaultAction]] = None


class Diagram(BaseModel):
    ipf: Any

    def _check_snapshot_id(self, snapshot_id):
        snapshot_id = snapshot_id or self.ipf.snapshot_id
        if snapshot_id not in self.ipf.loaded_snapshots:
            raise ValueError(f"Snapshot {snapshot_id} is not loaded or not found in IP Fabric.")
        return snapshot_id

    def _intent_overlay(self, overlay: Overlay) -> dict:
        try:
            Overlay.valid_intent_rule(overlay.intentRuleId)
        except ValueError:
            if not self.ipf.intent.loaded:
                self.ipf.intent.load_intent()
            intents = self.ipf.intent.intents_by_name[overlay.intentRuleId]
            if len(intents) > 1:
                raise ValueError(f"Multiple Intents found with name `{overlay.intentRuleId}`.")
            else:
                overlay.intentRuleId = intents[0].intent_id
        return overlay.model_dump()

    def _snapshot_overlay(self, overlay: Overlay, snapshot_id: str = None) -> dict:
        if overlay.snapshotToCompare not in self.ipf.loaded_snapshots:
            raise ValueError(f"Snapshot `{overlay.snapshotToCompare}` is not loaded.")
        overlay.snapshotToCompare = self.ipf.snapshots[overlay.snapshotToCompare].snapshot_id
        if snapshot_id or self.ipf.snapshot_id == overlay.snapshotToCompare:
            raise ValueError(f"Cannot compare snapshot `{overlay.snapshotToCompare}` to itself.")
        self._check_snapshot_id(overlay.snapshotToCompare)
        return overlay.model_dump()

    def _format_overlay(self, overlay: Union[Overlay, dict], snapshot_id: str = None) -> dict:
        if isinstance(overlay, dict):
            overlay = Overlay(**overlay)
        if overlay.intentRuleId:
            return self._intent_overlay(overlay)
        return self._snapshot_overlay(overlay, snapshot_id)

    def _query(
        self,
        parameters: Union[GRAPH_TYPES, dict],
        snapshot_id: Optional[str] = None,
        overlay: OVERLAY_SETTINGS = None,
        image: Literal["png", "svg", "json", "vsdx"] = "json",
        graph_settings: Union[GRAPH_SETTINGS, dict] = None,
        attr_filters: ATTRIBUTE_FILTERS = None,
        positions: POSITION_SETTINGS = None,
        api_version: Optional[Union[str, int]] = None,
    ) -> Union[dict, bytes]:
        """
        Submits a query, does no formatting on the parameters.  Use for copy/pasting from the webpage.
        :param parameters: dict: Dictionary to submit in POST.
        :return: list: List of Dictionary objects.
        """
        url = GRAPHS_URL
        if image in ["svg", "png", "vsdx"]:
            url += image
        payload = {
            "parameters": parameters if isinstance(parameters, dict) else parameters.model_dump(),
            "snapshot": self._check_snapshot_id(snapshot_id),
        }
        if overlay:
            payload["overlay"] = self._format_overlay(overlay, snapshot_id)
        if graph_settings:
            payload["settings"] = graph_settings if isinstance(graph_settings, dict) else graph_settings.model_dump()
        if attr_filters or self.ipf.attribute_filters:
            payload["attributeFilters"] = attr_filters or self.ipf.attribute_filters
        payload["positions"] = {k: dict(v) for k, v in positions.items()} if positions else {}
        if image == "vsdx":
            if payload.get("overlay", None):
                logger.warning("Overlay is not supported for Visio diagrams.")
            json_graph = raise_for_status(
                self.ipf.post(GRAPHS_URL, json=payload, headers=api_header(api_version))
            ).json()["graphResult"]
            files = {
                "graph": (
                    "graph.json",
                    json.dumps(
                        {
                            "graph": {
                                "boxLabels": json_graph["boxLabels"],
                                "edges": json_graph["graphData"]["edges"],
                                "nodes": json_graph["graphData"]["nodes"],
                            }
                        }
                    ),
                    "application/json",
                )
            }
            res = raise_for_status(self.ipf.post(url, files=files, headers=api_header(api_version)))
        else:
            res = raise_for_status(self.ipf.post(url, json=payload, headers=api_header(api_version)))
        return res.json() if image == "json" else res.content

    @staticmethod
    def _swap_src_dst(p: GRAPH_TYPES, unicast_swap_src_dst: bool = False) -> GRAPH_TYPES:
        return p.swap_src_dst if unicast_swap_src_dst and isinstance(p, Unicast) else p

    def json(
        self,
        parameters: GRAPH_TYPES,
        snapshot_id: Optional[str] = None,
        overlay: OVERLAY_SETTINGS = None,
        graph_settings: GRAPH_SETTINGS = None,
        attr_filters: ATTRIBUTE_FILTERS = None,
        unicast_swap_src_dst: bool = False,
        positions: POSITION_SETTINGS = None,
        api_version: Optional[Union[str, int]] = None,
    ) -> dict:
        return self._query(
            self._swap_src_dst(parameters, unicast_swap_src_dst),
            snapshot_id=snapshot_id,
            image="json",
            overlay=overlay,
            attr_filters=attr_filters,
            graph_settings=graph_settings.model_dump() if graph_settings else None,
            positions=positions,
            api_version=api_version,
        )

    def svg(
        self,
        parameters: GRAPH_TYPES,
        snapshot_id: Optional[str] = None,
        overlay: OVERLAY_SETTINGS = None,
        graph_settings: GRAPH_SETTINGS = None,
        attr_filters: ATTRIBUTE_FILTERS = None,
        unicast_swap_src_dst: bool = False,
        positions: POSITION_SETTINGS = None,
        api_version: Optional[Union[str, int]] = None,
    ) -> bytes:
        return self._query(
            self._swap_src_dst(parameters, unicast_swap_src_dst),
            snapshot_id=snapshot_id,
            overlay=overlay,
            attr_filters=attr_filters,
            image="svg",
            graph_settings=graph_settings.model_dump() if graph_settings else None,
            positions=positions,
            api_version=api_version,
        )

    def png(
        self,
        parameters: GRAPH_TYPES,
        snapshot_id: Optional[str] = None,
        overlay: OVERLAY_SETTINGS = None,
        graph_settings: GRAPH_SETTINGS = None,
        attr_filters: ATTRIBUTE_FILTERS = None,
        unicast_swap_src_dst: bool = False,
        positions: POSITION_SETTINGS = None,
        api_version: Optional[Union[str, int]] = None,
    ) -> bytes:
        return self._query(
            self._swap_src_dst(parameters, unicast_swap_src_dst),
            snapshot_id=snapshot_id,
            overlay=overlay,
            attr_filters=attr_filters,
            image="png",
            graph_settings=graph_settings.model_dump() if graph_settings else None,
            positions=positions,
            api_version=api_version,
        )

    def visio(
        self,
        parameters: GRAPH_TYPES,
        snapshot_id: Optional[str] = None,
        overlay: OVERLAY_SETTINGS = None,
        graph_settings: GRAPH_SETTINGS = None,
        attr_filters: ATTRIBUTE_FILTERS = None,
        unicast_swap_src_dst: bool = False,
        positions: POSITION_SETTINGS = None,
        api_version: Optional[Union[str, int]] = None,
    ) -> bytes:
        return self._query(
            self._swap_src_dst(parameters, unicast_swap_src_dst),
            snapshot_id=snapshot_id,
            overlay=overlay,
            attr_filters=attr_filters,
            image="vsdx",
            graph_settings=graph_settings.model_dump() if graph_settings else None,
            positions=positions,
            api_version=api_version,
        )

    def share_link(
        self,
        parameters: GRAPH_TYPES,
        snapshot_id: Optional[str] = None,
        overlay: OVERLAY_SETTINGS = None,
        graph_settings: GRAPH_SETTINGS = None,
        attr_filters: ATTRIBUTE_FILTERS = None,
        unicast_swap_src_dst: bool = False,
        positions: POSITION_SETTINGS = None,
    ) -> str:
        parameters = self._swap_src_dst(parameters, unicast_swap_src_dst)
        resp = self._query(
            parameters,
            snapshot_id=snapshot_id,
            overlay=overlay,
            image="json",
            attr_filters=attr_filters,
            graph_settings=graph_settings.model_dump() if graph_settings else None,
            positions=positions,
        )

        input_model = parameters.model_dump()
        input_model.pop("layouts", None)
        payload = {
            "graphView": {
                "name": "Shared view",
                "parameters": input_model,
                "collapsedNodeGroups": [],
                "hiddenNodes": [],
                "positions": {k: v["position"] for k, v in resp["graphResult"]["graphData"]["nodes"].items()},
                "settings": resp["graphResult"]["settings"],
            },
            "snapshot": self._check_snapshot_id(snapshot_id),
        }
        if overlay:
            payload["graphView"]["overlay"] = self._format_overlay(overlay, snapshot_id)
        res = raise_for_status(self.ipf.post("graphs/urls", json=payload))
        return urljoin(self.ipf.base_url, f"/diagrams/share/{res.json()['id']}")

    def model(
        self,
        parameters: GRAPH_TYPES,
        snapshot_id: Optional[str] = None,
        overlay: OVERLAY_SETTINGS = None,
        graph_settings: GRAPH_SETTINGS = None,
        attr_filters: ATTRIBUTE_FILTERS = None,
        unicast_swap_src_dst: bool = False,
        positions: POSITION_SETTINGS = None,
    ) -> GRAPH_RESULTS:
        json_data = self.json(
            parameters, snapshot_id, overlay, graph_settings, attr_filters, unicast_swap_src_dst, positions
        )
        graph_result = TopologyResult(**json_data) if "topology" in json_data else PathLookupResult(**json_data)
        e_setting = self._diagram_edge_settings(graph_result.graphResult.settings)

        edges = {edge_id: edge for edge_id, edge in graph_result.edges.items()}  # noqa: S7500
        for edge_id, edge in edges.items():
            edge.protocol = e_setting[edge.edgeSettingsId].name if edge.edgeSettingsId in e_setting else None
            if edge.source:
                edge.source = graph_result.nodes[edge.source]
            if edge.target:
                edge.target = graph_result.nodes[edge.target]

        graph_result.graphResult.graphData.edges = edges
        return graph_result if isinstance(graph_result, TopologyResult) else self._diagram_pathlookup(graph_result)

    @staticmethod
    def _diagram_pathlookup(graph_result: PathLookupResult) -> PathLookupResult:
        for edge_id, edge in graph_result.edges.items():
            for prev_id in edge.prevEdgeIds:
                edge.prevEdge.append(graph_result.edges[prev_id])
            for next_id in edge.nextEdgeIds:
                edge.nextEdge.append(graph_result.edges[next_id] if next_id in graph_result.edges else next_id)
        return graph_result

    @staticmethod
    def _diagram_edge_settings(graph_settings: Union[NetworkSettings, PathLookupSettings]) -> dict:
        edge_setting_dict = {}
        for edge in graph_settings.edges:
            edge_setting_dict[edge.id] = edge
            if isinstance(edge, GroupSettings):
                for child in edge.children:
                    edge_setting_dict[child.id] = child
        return edge_setting_dict

    def shared_view(
        self,
        url: Union[int, str],
        image: Literal["json", "code", "model", "svg", "png"] = "json",
        positions: bool = False,
    ):
        """Takes a shared graph link and returns the data or the code to implement in python.

        Args:
            url: Id of the shared view (1453653298) or full/partial URL (`/diagrams/share/1453626097`)
            image: Defaults to return the data instead of printing the code
            positions: If returning code then include positions of nodes in example; Default False

        Returns: The graph data or string representing the code to produce it.
        """
        query, _ = self.ipf._shared_url(url, False)
        if query["snapshot"] not in self.ipf.loaded_snapshots:
            logger.warning(f'Snapshot {query["snapshot"]} is not loaded, switching to {self.ipf.snapshot_id}.')
            query["snapshot"] = self.ipf.snapshot_id
        overlay = query["graphView"].get("overlay", {})
        if (
            overlay
            and overlay.get("snapshotToCompare", None)
            and overlay["snapshotToCompare"] not in self.ipf.loaded_snapshots
        ):
            logger.error(f"Snapshot `{overlay['snapshotToCompare']}` is not loaded to compare, removing overlay.")
            overlay = {}

        view = SharedView(
            snapshot_id=query["snapshot"],
            client_snapshot_id=self.ipf.snapshot_id,
            hidden_nodes=query["graphView"]["hiddenNodes"],
            collapsed_node_groups=query["graphView"]["collapsedNodeGroups"],
            positions=query["graphView"]["positions"],
            settings=query["graphView"]["settings"]["edges"],
            params=query["graphView"]["parameters"],
            hidden_devs=query["graphView"]["settings"].get("hiddenDeviceTypes", []),
            path_lookup=query["graphView"]["settings"].get("pathLookup", None),
            overlay=overlay,
        )
        if view.hidden_nodes or view.collapsed_node_groups:
            logger.warning("Hidden Nodes and Collapsed Node Groups are only available via the UI.")

        if image in ["json", "svg", "png"]:
            return self._query(
                parameters=query["graphView"]["parameters"],
                snapshot_id=query["snapshot"],
                overlay=overlay,
                image=image,
                graph_settings=query["graphView"]["settings"],
                positions=query["graphView"]["positions"],
            )

        if image == "code":
            return "\n".join(view.create_code(positions))
        if image == "model":
            return self.model(
                parameters=view.graph_model(),
                snapshot_id=query["snapshot"],
                overlay=overlay,
                graph_settings=view.graph_settings(),
                positions=query["graphView"]["positions"],
            )
        return None

    def _process_path(
        self, path: Union[GRAPH_RESULTS, GRAPH_TYPES], snapshot_id: Optional[str] = None
    ) -> GRAPH_RESULTS:
        """Helper method to process model or return a GraphResult.

        Args:
            path: PathLookupResult or TopologyResult or Unicast, Multicast, Host2GW, Network object
            snapshot_id: Optional snapshot_id to use

        Returns:
            GraphResult: Processed result
        """
        if isinstance(path, (PathLookupResult, TopologyResult)):
            return path
        return self.model(path, snapshot_id=snapshot_id)

    def nodes_in_diagram(
        self,
        path: Union[GRAPH_RESULTS, GRAPH_TYPES],
        dev_types: Optional[list[str]] = None,
        inventory_columns: Optional[Union[list[str], set[str]]] = None,
        snapshot_id: Optional[str] = None,
    ) -> list[Device]:
        """Fetches the device inventory for devices in the path.

        Args:
            path: PathLookupResult or TopologyResult or Unicast, Multicast, Host2GW, Network object
            dev_types: Optional list of device types to fetch, default ["l3switch", "fw", "switch", "router", "lb"]
            inventory_columns: Optional list of columns to fetch, default
                               ["sn", "version", "hostname", "model", "vendor", "siteName"]
            snapshot_id: Optional snapshot_id to use
        Returns:
            list[dict]: Device inventory
        """
        list_of_device_objects = []
        dev_types = NetworkSettings._valid_dev_types(dev_types or ["l3switch", "fw", "switch", "router", "lb"])
        inventory_columns = list(inventory_columns) or ["sn", "version", "hostname", "model", "vendor", "siteName"]
        if not all(_ in self.ipf.oas["tables/inventory/devices"].post.columns for _ in inventory_columns):
            raise ValueError(
                f"Inventory Columns '{inventory_columns}' must be in "
                f"{self.ipf.oas['tables/inventory/devices'].post.columns}."
            )

        path = self._process_path(path, snapshot_id)
        device_sn_to_fetch = [node.sn for node in path.nodes.values() if node.type in dev_types]
        for sn in device_sn_to_fetch:
            list_of_device_objects.append(self.ipf.devices.by_sn[sn])
        return list_of_device_objects

    def security_event_rules(  # NOSONAR
        self,
        path: Union[GraphResult, Unicast, Multicast, Host2GW],
        snapshot_id: Optional[str] = None,
    ) -> list[GraphDevice]:  # noqa: C901
        """Fetches the security events for devices in the path.

        Args:
            path: PathLookupResult or TopologyResult or Unicast, Multicast, Host2GW
            snapshot_id: Optional snapshot_id to use

        Returns:
            list[dict]: Device Inventory Information and Security Event
        """
        logger.warning("This method is experimental and may not work as expected.")
        path = self._process_path(path, snapshot_id)
        app_graph_nodes_dict = {node.id: node for node in path.nodes.values() if node.style}
        path_inventory_dict = {device.sn: device for device in self.nodes_in_diagram(path, snapshot_id)}
        list_for_return = []

        for vdevice, decisions in path.pathlookup.decisions.items():
            node = app_graph_nodes_dict.get(vdevice, None)
            if not node or not path_inventory_dict.get(node.sn, None):
                continue

            for trace in decisions.traces:
                for chain in trace.trace:
                    for event in chain.events:
                        if event.type in ["security check", "security check ignored"]:
                            device_data = path_inventory_dict[node.sn]
                            graph_device = GraphDevice(
                                device=device_data,
                                decidingPolicyName=event.decidingPolicyName,
                                decidingRule=event.decidingRule,
                                securityType=event.securityType,
                            )
                            if event.securityType == "zoneFw":
                                sec_model = graph_device.device.get_security_model()
                                machine_zones = sec_model.security.machineZones
                                rule_sequence = event.decidingRule[1]
                                zone_firewall_rules = (
                                    graph_device.device.technology.security.zone_firewall_policies.all(
                                        filters={
                                            "policyName": ["like", event.decidingPolicyName],
                                            "ruleNameNumeric": ["eq", (rule_sequence + 1)],
                                        }
                                    )
                                )
                                if zone_firewall_rules:
                                    graph_device.securityModelRule = zone_firewall_rules[0]
                                else:
                                    graph_device.securityModelRule = machine_zones.ruleChains.get(
                                        event.decidingPolicyName
                                    )
                            list_for_return.append(graph_device)
        return list_for_return

    def stp_instances(self, sites: Union[list[str], str], vlan_id: int) -> list[dict]:
        """Get the STP instances for the given sites and VLAN ID.

        Args:
            sites: Site name or list of site names to search for.
            vlan_id: VLAN ID to search for.

        Returns:

        """
        sites = [sites] if isinstance(sites, str) else sites
        filters = {"or": [{"siteName": ["eq", _]} for _ in sites]}
        devs = [_["sn"] for _ in self.ipf.inventory.devices.all(filters=filters, columns=["sn"])]
        roots = self.ipf.technology.stp.instance_members.all(
            filters={
                "and": [
                    {"members": ["containsAny", devs]},
                    {"rootHostname": ["empty", False]},
                    {"vlanId": ["eq", vlan_id]},
                ]
            },
            columns=["rootHostname", "rootId", "vlanId", "vlanName"],
        )
        return roots



import json
from functools import cached_property
from typing import Optional, Any, Union

import numpy as np
from ipfabric.models import Device
from pydantic import BaseModel, Field

from ipfabric.dynamic_attributes.configs import TableValue, ConfigValue
from ipfabric.dynamic_attributes.configs.rules import ConfigRule


CONFIG_MAP = {
    "current": {
        "no": "no_current_config",
        "config": "current_config",
        "name": "current",
    },
    "startup": {
        "no": "no_start_config",
        "config": "start_config",
        "name": "startup",
    },
    "log": {
        "no": "no_log_file",
        "config": "log_file",
        "name": "log_file",
    },
}


class BaseMatch(BaseModel):
    name: str = Field(description="The name of the rule that matched this device.")
    attribute: str = Field(description="The name of the attribute to assign.")
    current_value: Optional[str] = Field(None, description="The current global value of the attribute.")
    new_value: Optional[str] = Field(None, description="The new value to assign to the attribute.")
    mapping_match: Optional[dict] = Field(None, description="The mapping match from the mapping file.")
    regex_match: Optional[str] = Field(None, description="The original regex match.")
    delete: bool = Field(False, description="Whether to delete the attribute.")
    overwrite: bool = Field(False, description="Whether to overwrite existing attribute values.")

    @property
    def marked_for_deletion(self) -> bool:
        if self.delete and not self.new_value and self.current_value:
            return True
        return False

    def export(self) -> dict[str, Any]:
        return {
            "name": self.name,
            "new_value": self.new_value,
            "null_value": self.null_value,
            "regex_pattern": self.regex.pattern if self.regex else np.nan,
            "regex_group": self.regex.group if self.regex else np.nan,
            "regex_match": self.regex_match,
            "mapping_match": (json.dumps(self.mapping_match) if self.mapping_match else np.nan),
            "delete": self.delete,
            "overwrite": self.overwrite,
        }


class MatchedRule(BaseMatch, TableValue, BaseModel):
    column_value: Optional[str] = Field(None, description="The value of the column from the API response.")

    def export(self) -> dict[str, Any]:
        return {
            "api_endpoint": self.api_endpoint,
            "column_value": self.column_value if self.column else np.nan,
            "static": self.static if self.static else np.nan,
            **super().export(),
        }


class MatchedConfig(BaseMatch, ConfigValue, BaseModel):
    no_config_match: bool = Field(False, description="Whether the config was not found or supported.")

    def export(self) -> dict[str, Any]:
        return {
            "config": self.config,
            **super().export(),
        }


class Attribute(BaseModel):
    device: Device = Field(description="IP Fabric Device object.")
    matched_rule: dict[str, Union[MatchedRule, None]] = Field(
        default_factory=dict, description="Dictionary of Attributes mapped to the Matched TableRule."
    )
    no_start_config: Optional[bool] = Field(None, description="Whether the startup config is not supported.")
    no_current_config: Optional[bool] = Field(None, description="Whether the current config is not supported.")
    no_log_file: Optional[bool] = Field(None, description="Whether the log file is not supported.")
    start_config: Optional[str] = Field(None, description="Cached Startup Configuration.")
    current_config: Optional[str] = Field(None, description="Cached Current Configuration.")
    log_file: Optional[str] = Field(None, description="Cached Current Configuration.")

    @cached_property
    def sn(self) -> str:
        return self.device.sn

    def config_skip(self, rule: ConfigRule) -> bool:
        attribute = self.matched_rule.get(rule.attribute)
        if attribute and (not attribute.no_config_match or attribute.marked_for_deletion):
            # If an Attribute was set or no match was found, skip the device
            return True
        if getattr(self, CONFIG_MAP[rule.value.config]["no"]):
            # If the config is not supported, skip the device
            return True
        return False

    def global_attribute(self, name) -> Union[str, None]:
        return self.device.global_attributes.get(name)

    def local_attribute(self, name) -> Union[str, None]:
        return self.device.local_attributes.get(name)

    def export(self, report_columns: Union[None, list[str]] = None) -> list[dict[str, Any]]:
        matches = []
        for attribute, match in self.matched_rule.items():
            data = {
                **{_: self.device.model_dump(by_alias=True).get(_) for _ in report_columns},
                "attribute": attribute,
                "global_value": self.global_attribute(attribute),
                "local_value": self.local_attribute(attribute),
                "no_start_config": self.no_start_config,
                "no_current_config": self.no_current_config,
                "no_log_file": self.no_log_file,
            }
            if match:
                data.update(match.export())
            matches.append(data)
        return matches



from contextlib import asynccontextmanager
from pathlib import Path
from typing import Optional, Annotated

from dotenv import find_dotenv

try:
    import uvicorn
    from fastapi import FastAPI, Header, HTTPException, Request, BackgroundTasks, Depends, Query, Response
    from fastapi.responses import RedirectResponse
    from sqlmodel import Session, SQLModel, create_engine, select
except ImportError:
    raise ImportError(
        "fastapi, uvicorn, and sqlmodel are required to run this script. "
        "Please install them with 'pip3 install ipfabric[webhook]."
    )

from ipfabric.models import WebhookEvent
from ipfabric.scripts.webhook import (
    calc_method,
    query_db,
    verify_signature,
    WebhookConfig,
    WebhookResponse,
    AnnotatedWebhookEvent,
)

find_dotenv(usecwd=True) or Path("~").expanduser().joinpath(".env")
SETTINGS = WebhookConfig()
ENGINE = create_engine(SETTINGS.database_url, connect_args={"check_same_thread": False})


def get_session():
    with Session(ENGINE) as session:
        yield session


SessionDep = Annotated[Session, Depends(get_session)]


@asynccontextmanager
async def lifespan(app: FastAPI):
    SQLModel.metadata.create_all(ENGINE)
    yield


app = FastAPI(title=SETTINGS.title, lifespan=lifespan)


@app.get("/", response_class=RedirectResponse)
async def redirect_root():
    """Redirect to API docs"""
    return "/docs"


@app.get("/healthcheck", response_model=str)
async def healthcheck():
    """Simple healthcheck endpoint"""
    return "Ok"


@app.get("/snapshots/")
def read_snapshots(
    session: SessionDep,
    offset: int = 0,
    limit: Annotated[int, Query(le=1000)] = 100,
    snapshotId: Optional[str] = None,
) -> WebhookResponse:
    return WebhookResponse(**query_db(session, "snapshots", offset, limit, snapshotId))


@app.get("/intents/")
def read_intents(
    session: SessionDep,
    offset: int = 0,
    limit: Annotated[int, Query(le=1000)] = 100,
    snapshotId: Optional[str] = None,
) -> WebhookResponse:
    return WebhookResponse(**query_db(session, "intents", offset, limit, snapshotId))


@app.get("/events/")
def read_events(
    session: SessionDep,
    offset: int = 0,
    limit: Annotated[int, Query(le=1000)] = 100,
    snapshotId: Optional[str] = None,
) -> WebhookResponse:
    return WebhookResponse(**query_db(session, "events", offset, limit, snapshotId))


@app.get("/events/{dbId}")
def read_event(
    dbId: int,
    session: SessionDep,
) -> WebhookEvent:
    _ = session.exec(select(WebhookEvent).where(WebhookEvent.dbId == dbId)).first()
    return _ or Response(content="Not Found", status_code=404)


@app.post(SETTINGS.webhook_path)
async def webhook(
    event: AnnotatedWebhookEvent,
    request: Request,
    bg_tasks: BackgroundTasks,
    session: SessionDep,
    x_ipf_signature: str = Header(None),
):
    """IP Fabric Webhook Endpoint"""
    if not (_ := await verify_signature(request, x_ipf_signature, SETTINGS.ipf_secret))[0]:
        raise HTTPException(status_code=401, detail=_[1])

    return calc_method(event, bg_tasks, session, SETTINGS)


def main():
    uvicorn.run(
        "ipfabric.scripts.webhook_api:app",
        host=SETTINGS.uvicorn_ip,
        port=SETTINGS.uvicorn_port,
        reload=SETTINGS.uvicorn_reload,
    )


if __name__ == "__main__":
    main()



import json
from functools import cached_property
from typing import Optional, Literal, Any, Union, Callable, ForwardRef

import yaml
import yaml_include
from pydantic import BaseModel, Field, model_validator, ConfigDict, FilePath

from ipfabric.dynamic_attributes.configs.rules import (
    InventoryRule,
    DefaultRule,
    ConfigRule,
    TableRule,
    DefaultConfigRule,
)

try:
    import tomllib
except ImportError:
    import tomli as tomllib

Config = ForwardRef("Config")


class IPFabric(BaseModel):
    base_url: Optional[str] = Field(
        None,
        description="The IP Fabric Base URL to fetch data from (env: 'IPF_URL').",
        title="IP Fabric URL",
        examples=["https://demo.ipfabric.com"],
    )
    auth: Optional[str] = Field(
        None,
        description="The IP Fabric API token to use for authentication (env: 'IPF_TOKEN'). "
        "Username and password can be used by setting Environment Variables (IPF_USERNAME, IPF_PASSWORD).",
        title="IP Fabric API Token",
    )
    timeout: Optional[Union[int, tuple, float, None]] = Field(
        5,
        description="The timeout for the API requests; default 5 seconds (env: 'IPF_TIMEOUT').",
        title="IP Fabric Timeout",
    )
    verify: Union[bool, str] = Field(
        True, description="Verify SSL Certificates; default True (env: 'IPF_VERIFY').", title="SSL Verification"
    )
    snapshot_id: str = Field(
        "$last",
        description="The snapshot ID to use for the API endpoint; defaults to '$last'.",
        title="Snapshot ID",
        examples=["$last", "$prev", "$lastLocked", "d03a89d3-911b-4e2d-868b-8b8103771801"],
    )


class Config(BaseModel):  # noqa: F811
    model_config = ConfigDict(extra="allow")

    ipfabric: Optional[IPFabric] = Field(
        default_factory=IPFabric, description="IP Fabric connection configuration.", title="IP Fabric Connection"
    )
    dry_run: bool = Field(
        True, description="Defaults to run in dry-run mode and not apply any updates.", title="Dry Run"
    )
    update_snapshot: bool = Field(
        True,
        description="Update Local Attributes on the selected snapshot; default True.",
        title="Update Snapshot Attributes",
    )
    inventory: InventoryRule = Field(
        default_factory=InventoryRule,
        description="Optional: Filters to limit the inventory of devices based on Inventory > Devices table.",
        title="Inventory Filters",
    )
    default: Optional[DefaultRule] = Field(
        None,
        description="Optional: Default configuration to merge into all other Table Rules.",
        title="Default Table TableRule",
    )
    default_config: Optional[DefaultConfigRule] = Field(
        None,
        description="Optional: Default configuration to merge into all other Configuration Rules.",
        title="Default Configuration TableRule",
    )
    rules: list[Union[ConfigRule, TableRule]] = Field(
        description="List of Table or Configuration Rules which are processed in order; at least 1 rule is required.",
        title="Dynamic Attribute Rules",
    )

    @model_validator(mode="after")
    def _validate(self):
        if not self.rules:
            raise ValueError("At least one rule must be provided.")
        if len({_.name for _ in self.rules}) != len(self.rules):
            raise ValueError("Duplicate TableRule Names found.")
        if False in {bool(_.value) for _ in self.rules} and not self.default.value:
            raise ValueError("All Rules must have a value set or 'default[value]' can be used for Table Rules.")
        if False in {bool(_.attribute) for _ in self.rules if isinstance(_, TableRule)} and not self.default.attribute:
            raise ValueError(
                "An Attribute Name must be set in 'default[attribute]' or all Table rules must have it defined."
            )
        if (
            False in {bool(_.attribute) for _ in self.rules if isinstance(_, ConfigRule)}
            and not self.default_config.attribute
        ):
            raise ValueError(
                "An Attribute Name must be set in 'default_config[attribute]' "
                "or all Configuration Rules must have it defined."
            )
        return self

    @cached_property
    def merged_rules(self) -> list[Union[ConfigRule, TableRule]]:
        """Copy Defaults to rules"""
        return [
            rule.merge_default(self.default if isinstance(rule, TableRule) else self.default_config)
            for rule in self.rules
        ]

    def model_dump_merged(
        self,
        *,
        mode: Literal["json", "python"] = "python",
        context: Optional[Any] = None,
        by_alias: Optional[bool] = None,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        round_trip: bool = False,
        warnings: Union[bool, Literal["none", "warn", "error"]] = True,
        fallback: Optional[Callable[[Any], Any]] = None,
        serialize_as_any: bool = False,
    ) -> dict[str, Any]:
        """Dump the full config with all default rules merged."""
        config = self.model_copy(deep=True)
        config.rules = self.merged_rules
        return config.model_dump(
            mode=mode,
            exclude={"default", "default_config"},
            context=context,
            by_alias=by_alias,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
            round_trip=round_trip,
            warnings=warnings,
            fallback=fallback,
            serialize_as_any=serialize_as_any,
        )

    def model_merged(self) -> "Config":
        """Return a copy of the merged config with all default rules merged."""
        config = self.model_copy(deep=True)
        config.rules = self.merged_rules
        config.default, config.default_config = None, None
        return config


Config.model_rebuild()


def my_loader(urlpath, file, Loader):
    if urlpath.lower().endswith(".json"):
        return json.load(file)
    if urlpath.lower().endswith(".toml"):
        return tomllib.load(file)
    if urlpath.lower().endswith((".yaml", ".yml")):
        return yaml.load(file, Loader)
    raise ValueError("Unsupported config file format. Use .yaml, .yml, .json, or .toml.")


def load_config(file: FilePath) -> Config:
    try:
        if (suffix := file.suffix.lower()) in (".yaml", ".yml"):
            yaml.add_constructor("!inc", yaml_include.Constructor(custom_loader=my_loader), Loader=yaml.SafeLoader)
            with open(file, "r", encoding="utf8") as f:
                return Config(**yaml.safe_load(f), filename=file.name)
        elif suffix == ".json":
            with open(file, "r", encoding="utf8") as f:
                return Config(**json.load(f), filename=file.name)
        elif suffix == ".toml":
            with open(file, "rb") as f:
                return Config(**tomllib.load(f), filename=file.name)
        else:
            raise ValueError("Unsupported config file format. Use .yaml, .yml, .json, or .toml.")
    except (yaml.YAMLError, json.JSONDecodeError, tomllib.TOMLDecodeError) as exc:
        raise type(exc)(f"Error parsing {str(file)} file: {exc}")


class MultiConfig(BaseModel):
    """Model to handle multiple configurations."""

    configs: list[Union[Config, FilePath]] = Field(
        title="IP Fabric Dynamic Attribute Configurations",
        description="Ordered list of Configurations to process; at least one configuration is required. "
        "IPFabric ['base_url', 'auth', 'verify', 'snapshot_id'] values must be the same in all config files; "
        "recommended to leave empty and use environment variables.",
    )
    dry_run_override: Optional[bool] = Field(
        None,
        description="Defaults to use the 'dry_run' value from the configs; values in all the configs must be the same. "
        "If set to a boolean value then it will override the config values.",
        title="Override Dry Run",
    )
    update_snapshot_override: Optional[bool] = Field(
        None,
        description="Defaults to use the 'update_snapshot' value from the configs; "
        "values in all the configs must be the same. "
        "If set to a boolean value then it will override the config values.",
        title="Update Snapshot Attributes",
    )
    ipfabric_override: Optional[IPFabric] = Field(
        None,
        description="Defaults to use the 'ipfabric' value from the configs; "
        "values in all the configs must be the same. "
        "If an IP Fabric configuration is passed in then the value will ignore the individual the configs.",
        title="IP Fabric Connection",
    )

    def _get_ipfabric_config(self, attr: str) -> list[Any]:
        """Get a unique attribute from all configurations."""
        return list(
            {getattr(_.ipfabric, attr) for _ in self.configs if _.ipfabric and getattr(_.ipfabric, attr) is not None}
        )

    @model_validator(mode="after")
    def _validate(self):
        if not self.configs:
            raise ValueError("At least one configuration must be provided.")
        self.configs = [_ if isinstance(_, Config) else load_config(_) for _ in self.configs]
        if not self.ipfabric_override:
            for attr in ["base_url", "auth", "verify", "snapshot_id"]:
                if len(self._get_ipfabric_config(attr)) > 1:
                    raise ValueError(
                        "All IPFabric settings must have the same ['base_url', 'auth', 'verify', 'snapshot_id'] values."
                        " Alternatively, you can set 'ipfabric_override' to a single IPFabric configuration."
                    )
        if self.dry_run_override is None and len({_.dry_run for _ in self.configs}) > 1:
            raise ValueError(
                "All configurations must have the same 'dry_run' value. "
                "Alternatively, you can set 'dry_run_override' to a boolean value."
            )
        if self.update_snapshot_override is None and len({_.update_snapshot for _ in self.configs}) > 1:
            raise ValueError(
                "All configurations must have the same 'update_snapshot' value. "
                "Alternatively, you can set 'update_snapshot_override' to a boolean value."
            )
        return self

    @staticmethod
    def _merge_timeout(timeouts: set) -> Union[int, tuple, float, None]:
        """Merge timeouts from a set of timeouts using the largest value or None."""
        if None in timeouts or 0 in timeouts:
            return None
        elif t := {_ for _ in timeouts if isinstance(_, (int, float))}:
            return sorted(t, reverse=True)[0]
        return sorted(timeouts)[0]

    def _merge_ipf_settings(self) -> IPFabric:
        """Merge IPFabric settings."""
        base_url = self._get_ipfabric_config("base_url")
        auth = self._get_ipfabric_config("auth")
        verify = self._get_ipfabric_config("verify")
        snapshot_id = self._get_ipfabric_config("snapshot_id")
        return IPFabric(
            base_url=base_url[0] if base_url else None,
            auth=auth[0] if auth else None,
            verify=verify[0] if verify else None,
            timeout=self._merge_timeout(set(self._get_ipfabric_config("timeout"))),
            snapshot_id=snapshot_id[0] if snapshot_id else None,
        )

    def merge_configs(self) -> Config:
        """Merge all configurations into a single config file."""
        rules = []
        reports = set()
        for idx, config in enumerate(self.configs):
            prefix = config.model_extra.get("filename", f"config_{idx}")
            merged = config.model_merged()
            reports.update(set(merged.inventory.report_columns))
            for rule in merged.rules:
                rule.name = f"{prefix}: {rule.name}"
                rule.inventory = merged.inventory
            rules.extend(merged.rules)

        return Config(
            ipfabric=self.ipfabric_override or self._merge_ipf_settings(),
            dry_run=self.configs[0].dry_run if self.dry_run_override is None else self.dry_run_override,
            update_snapshot=(
                self.configs[0].update_snapshot
                if self.update_snapshot_override is None
                else self.update_snapshot_override
            ),
            inventory=InventoryRule(report_columns=list(reports)),
            rules=rules,
        )
